@@ -18,14 +18,30 @@
 
 #include "buildid.h"
 
+/*
+ * Allow the build system to generate version.h (and define
+ * the HAVE_VERSION_H preprocessor macro) or get the version via the BUILD_ID
+ * preprocessor macro.  If neither is available, use a sensible default.
+ */
+#ifdef HAVE_VERSION_H
+#include "version.h"
+#elif defined(BUILD_ID)
+#define STR(x) #x
+#define XSTR(x) STR(x)
+#define VERSION_STRING XSTR(BUILD_ID)
+#endif
+#ifndef VERSION_STRING
+#define VERSION_STRING "4.2.6"
+#endif
+
 const char *buildid = VERSION_NAME " " VERSION_STRING;
 const char *buildver = VERSION_STRING;
 
 /**
- * Hack -- Link a copyright message into the executable
+ * Link a copyright message into the executable
  */
 const char *copyright =
-	"Copyright (c) 1987-2020 Angband contributors.\n"
+	"Copyright (c) 1987-2022 Angband contributors.\n"
 	"\n"
 	"This work is free software; you can redistribute it and/or modify it\n"
 	"under the terms of either:\n"
