@@ -245,10 +245,7 @@ static struct object *rd_item(void)
 	}
 
 	/* Set effect */
-	if (effect && obj->ego)
-		obj->effect = obj->ego->effect;
-
-	if (effect && !obj->effect)
+	if (effect)
 		obj->effect = obj->kind->effect;
 
 	/* Success */
@@ -349,7 +346,7 @@ static bool rd_monster(struct chunk *c, struct monster *mon)
 
 	/* Now delete the monster if necessary */
 	if (delete) {
-		delete_monster(mon->grid);
+		delete_monster(c, mon->grid);
 	}
 
 	return true;
@@ -784,14 +781,14 @@ int rd_player(void)
 	rd_s16b(&player->max_depth);
 	rd_s16b(&player->recall_depth);
 
-	/* Hack -- Repair maximum player level */
+	/* Repair maximum player level */
 	if (player->max_lev < player->lev) player->max_lev = player->lev;
 
-	/* Hack -- Repair maximum dungeon level */
+	/* Repair maximum dungeon level */
 	if (player->max_depth < 0) player->max_depth = 1;
 	if (player->recall_depth <= 0) player->recall_depth = player->max_depth;
 
-	/* Hack -- Reset cause of death */
+	/* Reset cause of death */
 	if (player->chp >= 0)
 		my_strcpy(player->died_from, "(alive and well)",
 				  sizeof(player->died_from));
@@ -801,7 +798,8 @@ int rd_player(void)
 	player->old_grid.y = tmp8u;
 	rd_byte(&tmp8u);
 	player->old_grid.x = tmp8u;
-	strip_bytes(5);
+	strip_bytes(4);
+	rd_byte(&player->skip_cmd_coercion);
 	rd_byte(&player->unignoring);
 	rd_s16b(&player->deep_descent);
 
@@ -1180,7 +1178,8 @@ int rd_gear(void)
 	for (obj = player->gear, known_obj = player->gear_k; obj;
 		 obj = obj->next, known_obj = known_obj->next) {
 		obj->known = known_obj;
-		player->upkeep->total_weight += (obj->number * obj->weight);
+		player->upkeep->total_weight +=
+			obj->number * object_weight_one(obj);
 	}
 
 	calc_inventory(player);
@@ -1199,9 +1198,14 @@ static int rd_stores_aux(rd_item_t rd_item_version)
 
 	/* Read the stores */
 	rd_u16b(&tmp16u);
+	if (tmp16u != z_info->store_max) {
+		note(format("The number of stores in the savefile (%u) is "
+			"different than expected (%u).", tmp16u,
+			z_info->store_max));
+	}
 	for (i = 0; i < tmp16u; i++) {
-		struct store *store = &stores[i];
-
+		struct store *store = (i < z_info->store_max) ?
+			 &stores[i] : NULL;
 		uint8_t own, num;
 
 		/* Read the basic info */
@@ -1209,7 +1213,9 @@ static int rd_stores_aux(rd_item_t rd_item_version)
 		rd_byte(&num);
 
 		/* XXX: refactor into store.c */
-		store->owner = store_ownerbyidx(store, own);
+		if (store) {
+			store->owner = store_ownerbyidx(store, own);
+		}
 
 		/* Read the items */
 		for (; num; num--) {
@@ -1229,11 +1235,19 @@ static int rd_stores_aux(rd_item_t rd_item_version)
 			obj->known = known_obj;
 
 			/* Accept any valid items */
-			if (store->stock_num < z_info->store_inven_max && obj->kind) {
-				if (store->sidx == STORE_HOME)
+			if (store && store->stock_num
+					< z_info->store_inven_max
+					&& obj->kind) {
+				if (store->feat == FEAT_HOME) {
 					home_carry(obj);
-				else
+				} else {
 					store_carry(store, obj);
+				}
+			} else {
+				if (obj->known) {
+					object_delete(NULL, NULL, &obj->known);
+				}
+				object_delete(NULL, NULL, &obj);
 			}
 		}
 	}
@@ -1610,7 +1624,7 @@ int rd_chunks(void)
 
 	rd_u16b(&chunk_max);
 	for (j = 0; j < chunk_max; j++) {
-		struct chunk *c;
+		struct chunk *c = NULL;
 
 		/* Read the dungeon */
 		if (rd_dungeon_aux(&c))
@@ -1652,7 +1666,7 @@ int rd_chunks(void)
 			rd_u16b(&tmp16u);
 			c->width = tmp16u;
 			rd_u16b(&c->feeling_squares);
-			for (i = 0; i < z_info->f_max + 1; i++) {
+			for (i = 0; i < FEAT_MAX + 1; i++) {
 				rd_u16b(&tmp16u);
 				c->feat_count[i] = tmp16u;
 			}
