@@ -49,8 +49,6 @@
 # include <curses.h>
 #endif
 
-#include <term.h>
-
 #undef term
 
 /**
@@ -60,8 +58,8 @@
 
 
 /**
- * Hack -- Windows Console mode uses PDCURSES and cannot do any terminal stuff
- * Hack -- Windows needs Sleep(), and I really don't want to pull in all
+ * Windows Console mode uses PDCURSES and cannot do any terminal stuff
+ * Windows needs Sleep(), and I really don't want to pull in all
  *         the Win32 headers for this one function
  */
 #if defined(WIN32_CONSOLE_MODE)
@@ -101,12 +99,6 @@ static struct termios  game_termios;
 
 #endif
 
-/**
- * The TERM environment variable; used for terminal capabilities.
- */
-static char *termtype;
-static bool loaded_terminfo;
-
 /**
  * Simple rectangle type
  */
@@ -157,7 +149,7 @@ static int active = 0;
 #ifdef A_COLOR
 
 /**
- * Hack -- define "A_BRIGHT" to be "A_BOLD", because on many
+ * Define "A_BRIGHT" to be "A_BOLD", because on many
  * machines, "A_BRIGHT" produces ugly "inverse" video.
  */
 #ifndef A_BRIGHT
@@ -258,7 +250,7 @@ static void keymap_game_prepare(void) {
 	game_termios.c_cc[VSUSP] = (char)26;
 
 #ifdef VDSUSP
-	/* Hack -- disable "Ctrl-Y" on *BSD */
+	/* Disable "Ctrl-Y" on *BSD */
 	game_termios.c_cc[VDSUSP] = (char)-1;
 #endif
 
@@ -295,7 +287,7 @@ static errr Term_xtra_gcu_alive(int v) {
 		echo();
 		nl();
 
-		/* Hack -- make sure the cursor is visible */
+		/* Make sure the cursor is visible */
 		Term_xtra(TERM_XTRA_SHAPE, 1);
 
 		/* Flush the curses buffer */
@@ -329,21 +321,252 @@ static errr Term_xtra_gcu_alive(int v) {
 	return 0;
 }
 
+/**
+ * Usage:
+ *
+ * angband -mgcu
+ *
+ * or
+ *
+ * angband -mgcu -- [one or more options described in help_gcu below]
+ */
 const char help_gcu[] = "Text mode, subopts\n"
 	"              -B     Use brighter bold characters\n"
 	"              -D     Use terminal default background color\n"
 	"              -K     Keep terminal's color table when changing colors\n"
-	"              -nN    Use N terminals (up to 6)";
+	"              -nN    Use N terminals (up to 6, calculate size automatically. This option cannot be used with below options)\n"
+	"                   To manually set terminal sizes use below options\n"
+	"              -right (dimension)[,dimension]\n"
+	"              -left (dimension)[,dimension]\n"
+	"              -bottom (dimension)[,dimension]\n"
+	"              -top (dimension)[,dimension]\n"
+	"              -spacer (dimension)     space between terminals\n"
+	"                   Where dimension is = (columns)x(rows)\n"
+	"                   ex: 10x10\n"
+	"                   one of the value can also be ommited like 10x*\n"
+	"                   where * is automatically calculated if possible\n"
+	"                   if second dimension is specified then that side is split and second terminal is created.\n"
+	"                   incase of second dimension it can be automatically calculated by using *, ex: 10x10,*\n"
+	"                   second dimension is not allowed for top and bottom\n"
+	"                   NOTE: order of arguments matters\n"
+	"                   if top and bottom specified first then entire row is given to them, if they are specified after left, then margin is left for terminal and then the top and bottom terminal starts\n"
+	;
+
+#ifdef MSYS2_ENCODING_WORKAROUND
+/*
+ * Override the default character encoding in MSYS2 for better handling in
+ * Term_text_gcu() of characters that don't fall in the ASCII range.  See
+ * init_sdl2() in main-sdl2.c for more details about the default character
+ * encoding in MSYS2.
+ */
 
 /**
- * Usage:
+ * Convert UTF-8 to UTF-16 with each UTF-16 in the native byte order and
+ * lossily change any code point that requires a surrogate pair to U+FFFD.
+ * Return the total number of code points that would be generated by
+ * converting the UTF-8 input.
  *
- * angband -mgcu -- [-B] [-D] [-nN]
+ * \param dest Points to the buffer in which to store the conversion.  May be
+ * NULL.
+ * \param src Is a null-terminated UTF-8 sequence.
+ * \param n Is the maximum number of code points to store in dest.
  *
- *   -B      Use brighter bold characters
- *   -D      Use terminal default background color
- *   -nN     Use N terminals (up to 6)
+ * In case of malformed UTF-8, inserts a U+FFFD in the converted output at the
+ * point of the error.
  */
+static size_t Term_mbcs_gcu_msys2(wchar_t *dest, const char *src, int n)
+{
+        size_t nout = (n > 0) ? n : 0;
+        size_t count = 0;
+
+        while (1) {
+                /*
+                 * Default to U+FFFD to indicate an erroneous UTF-8 sequence
+                 * that could not be decoded.  Follow "best practice"
+                 * recommended by the Unicode 6 standard:  an erroneous
+		 * sequence ends as soon as a disallowed byte is encountered.
+                 */
+                uint32_t decoded = 0xfffd;
+
+                if (((unsigned int) *src & 0x80) == 0) {
+                        /*
+                         * Encoded as a single byte:  U+0000 to U+007F ->
+                         * 0xxxxxxx.
+                         */
+                        if (*src == 0) {
+                                if (dest && count < nout) {
+                                        dest[count] = 0;
+                                }
+                                break;
+                        }
+                        decoded = *src;
+                        ++src;
+                } else if (((unsigned int) *src & 0xe0) == 0xc0) {
+                        /*
+                         * Encoded as two bytes:  U+0080 to U+07FF ->
+                         * 110xxxxx 10xxxxxx.
+                         */
+                        uint32_t part = ((uint32_t) *src & 0x1f) << 6;
+
+                        ++src;
+                        /*
+                         * Check that the first two bits of the continuation
+                         * byte are valid and the encoding is not overlong.
+                         */
+                        if (((unsigned int) *src & 0xc0) == 0x80
+                                        && part > 0x40) {
+                                decoded = part + ((uint32_t) *src & 0x3f);
+                                ++src;
+                        }
+                } else if (((unsigned int) *src & 0xf0) == 0xe0) {
+                        /*
+                         * Encoded as three bytes:  U+0800 to U+FFFF ->
+                         * 1110xxxx 10xxxxxx 10xxxxxx.
+                         */
+                        uint32_t part = ((uint32_t) *src & 0xf) << 12;
+
+                        ++src;
+                        if (((unsigned int) *src & 0xc0) == 0x80) {
+                                part += ((uint32_t) *src & 0x3f) << 6;
+                                ++src;
+                                /*
+                                 * The second part of the test rejects
+                                 * overlong encodings.  The third part
+                                 * rejects encodings of U+D800 to U+DFFF,
+                                 * reserved for surrogate pairs.
+                                 */
+                                if (((unsigned int) *src & 0xc0) == 0x80
+                                                && part >= 0x800
+                                                && (part & 0xf800) != 0xd800) {
+                                        decoded = part
+                                                + ((uint32_t) *src & 0x3f);
+                                        ++src;
+                                }
+                        }
+                } else if (((unsigned int) *src & 0xf8) == 0xf0) {
+                        /*
+                         * Encoded as four bytes:  U+10000 to U+1FFFFF ->
+                         * 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx.
+                         */
+                        uint32_t part = ((uint32_t) *src & 0x7) << 18;
+
+                        ++src;
+                        if (((unsigned int) *src & 0xc0) == 0x80) {
+                                part += ((uint32_t) * src & 0x3f) << 12;
+                                ++src;
+                                /*
+                                 * The second part of the test rejects
+                                 * overlong encodings.  The third part
+                                 * rejects code points beyond U+10FFFF which
+                                 * can't be encoded in UTF-16.
+                                 */
+                                if (((unsigned int) *src & 0xc0) == 0x80
+                                                && part >= 0x10000
+                                                && (part & 0xff0000)
+                                                <= 0x100000) {
+                                        part += ((uint32_t) *src & 0x3f) << 6;
+                                        ++src;
+                                        if (((unsigned int) *src & 0xc0)
+                                                        == 0x80) {
+                                                decoded = part
+                                                        + ((uint32_t) *src
+                                                        & 0x3f);
+                                                ++src;
+                                        }
+                                }
+                        }
+                } else {
+                        /*
+                         * Either an impossible byte or one that signals the
+                         * start of a five byte or longer encoding.
+                         */
+                        ++src;
+                }
+                if (dest && count < nout) {
+                        if (decoded >= 0x10000) {
+                                /*
+                                 * Would require a surrogate pair to represent
+                                 * accurately.  Substitute U+FFFD instead.
+                                 */
+                                assert(decoded <= 0x10FFFF);
+                                dest[count] = (wchar_t) 0xfffd;
+                        } else {
+                                /*
+                                 * By the decoding logic, the decoded value
+                                 * should not match part of surrogate pair.
+                                 */
+                                assert(decoded < 0xd800 || decoded > 0xdfff);
+                                dest[count] = (wchar_t) decoded;
+                        }
+                }
+                ++count;
+        }
+	return count;
+}
+
+/**
+ * Convert a UTF-16 stored in the native byte order to UTF-8.
+ * \param s Points to the buffer where the conversion should be stored.
+ * That buffer must have at least Term_wcsz_gcu_msys2() bytes.
+ * \param wchar Is the UTF-16 value to convert.
+ * \return The returned value is the number of bytes written to s or -1
+ * if the UTF-16 value could not be converted.
+ *
+ * This is a necessary counterpart to Term_mbcs_gcu_msys2():  since we
+ * are overriding the default multibyte to wide character conversion, need
+ * to override the reverse conversion as well.
+ */
+static int Term_wctomb_gcu_msys2(char *s, wchar_t wchar)
+{
+        if ((unsigned int) wchar <= 0x7f) {
+                *s = wchar;
+                return 1;
+        }
+        if ((unsigned int) wchar < 0x7ff) {
+                *s++ = 0xc0 + (((unsigned int) wchar & 0x7c0) >> 6);
+                *s++ = 0x80 + ((unsigned int) wchar & 0x3f);
+                return 2;
+        }
+        /* Refuse to encode a value reserved for surrogate pairs in UTF-16. */
+        if ((unsigned int) wchar >= 0xd800 && (unsigned int) wchar <= 0xdfff) {
+                return -1;
+        }
+        *s++ = 0xe0 + (((unsigned int) wchar & 0xf000) >> 12);
+        *s++ = 0x80 + (((unsigned int) wchar & 0xfc0) >> 6);
+        *s++ = 0x80 + ((unsigned int) wchar & 0x3f);
+        return 3;
+}
+
+/**
+ * Return whether a UTF-16 value is printable.
+ *
+ * This is a necessary counterpart to Term_mbcs_gcu_msys2() so that
+ * screening of wide characters in the core's text_out_to_screen() is
+ * consistent with what Term_mbcs_gcu_msys2() does.
+ */
+static int Term_iswprint_gcu_msys2(wint_t wc)
+{
+       /*
+         * Upcast the UTF-16 value to UTF-32 since the UTF-16 value is either
+         * equal to the code point's value or is part of a surrogate pair.
+         */
+        return utf32_isprint((uint32_t) wc);
+}
+
+/**
+ * Return the maximum number of bytes needed for a multibyte encoding of a
+ * wchar.
+ */
+static int Term_wcsz_gcu_msys2(void)
+{
+        /*
+         * UTF-8 takes at most 3 bytes to encode a code point that can be
+         * encoded with a single 16-bit quantity in UTF-16.
+         */
+        return 3;
+}
+
+#endif /* MSYS2_ENCODING_WORKAROUND */
 
 /**
  * Init the "curses" system
@@ -360,7 +583,7 @@ static void Term_init_gcu(term *t) {
 	/* Count init's, handle first */
 	if (active++ != 0) return;
 
-	#if defined(USE_NCURSES) && defined(KEY_MOUSE)
+	#if defined(NCURSES_MOUSE_VERSION) && defined(KEY_MOUSE)
 	/* Turn on the mouse. */
 	mousemask(ALL_MOUSE_EVENTS, NULL);
 	#endif
@@ -376,6 +599,17 @@ static void Term_init_gcu(term *t) {
 
 	/* Game keymap */
 	keymap_game();
+
+#ifdef MSYS2_ENCODING_WORKAROUND
+	/*
+	 * Override default character encoding on MSYS2; see main-sdl2.c for
+	 * more details.
+	 */
+	text_mbcs_hook = Term_mbcs_gcu_msys2;
+	text_wctomb_hook = Term_wctomb_gcu_msys2;
+	text_wcsz_hook = Term_wcsz_gcu_msys2;
+	text_iswprint_hook = Term_iswprint_gcu_msys2;
+#endif /* MSYS2_ENCODING_WORKAROUND */
 }
 
 
@@ -392,7 +626,7 @@ static void Term_nuke_gcu(term *t) {
 	/* Count nuke's, handle last */
 	if (--active != 0) return;
 
-	/* Hack -- make sure the cursor is visible */
+	/* Make sure the cursor is visible */
 	Term_xtra(TERM_XTRA_SHAPE, 1);
 
 #ifdef A_COLOR
@@ -580,8 +814,12 @@ static errr Term_xtra_gcu_event(int v) {
 		if (i == EOF) return (1);
 	}
 
-	/* Not sure if this is portable to non-ncurses platforms */
-	#ifdef USE_NCURSES
+	/*
+	 * Both NCurses and PDCurses define KEY_RESIZE.  According to
+	 * PDCurses User's Guide, KEY_RESIZE is not part of the X/Open
+	 * specification so this generally will not work with Curses.
+	 */
+	#ifdef KEY_RESIZE
 	if (i == KEY_RESIZE) {
 		/* wait until we go one second (10 deci-seconds) before actually
 		 * doing the resizing. users often end up triggering multiple
@@ -596,7 +834,7 @@ static errr Term_xtra_gcu_event(int v) {
 	}
 	#endif
 
-	#if defined(USE_NCURSES) && defined(KEY_MOUSE)
+	#if defined(NCURSES_MOUSE_VERSION) && defined(KEY_MOUSE)
 	if (i == KEY_MOUSE) {
 		MEVENT m;
 		if (getmouse(&m) != OK) return (0);
@@ -773,7 +1011,7 @@ static void handle_extended_color_tables(void) {
 			for (i = 0; i < BASIC_COLORS; i++) {
 				/*
 				 * Scale components to a range of 0 - 1000 per
-				 * per init_color()'s documentation.
+				 * init_color()'s documentation.
 				 */
 				init_color(i,
 					(angband_color_table[i][1] * 1001) / 256,
@@ -822,8 +1060,10 @@ static errr Term_xtra_gcu(int n, int v) {
 		/* Clear screen */
 		case TERM_XTRA_CLEAR: touchwin(td->win); wclear(td->win); return 0;
 
-		/* Make a noise */
-		case TERM_XTRA_NOISE: write(1, "\007", 1); return 0;
+		/* Make a noise; beep() has been part of the Curses interface
+		 * since 1984; on systems not capable of an audible warning,
+		 * it may flash the screen */
+		case TERM_XTRA_NOISE: beep(); return 0;
 
 		/* Flush the Curses buffer */
 		case TERM_XTRA_FRESH: wrefresh(td->win); return 0;
@@ -866,7 +1106,7 @@ static errr Term_curs_gcu(int x, int y) {
 
 /**
  * Erase a grid of space
- * Hack -- try to be "semi-efficient".
+ * Try to be "semi-efficient".
  */
 static errr Term_wipe_gcu(int x, int y, int n) {
 	term_data *td = (term_data *)(Term->data);
@@ -909,7 +1149,7 @@ static errr Term_text_gcu(int x, int y, int n, int a, const wchar_t *s) {
 		int color;
 
 		/* Set bg and fg to the same color when drawing solid walls */
-		if (a / MAX_COLORS == BG_SAME) {
+		if (a / MULT_BG == BG_SAME) {
 			color = same_colortable[attr];
 		} else {
 			color = colortable[attr];
@@ -957,10 +1197,6 @@ static errr term_data_init_gcu(term_data *td, int rows, int cols, int y, int x)
 	/* Avoid bottom right corner */
 	t->icky_corner = true;
 
-	/* Erase with "white space" */
-	t->attr_blank = COLOUR_WHITE;
-	t->char_blank = ' ';
-
 	/* Differentiate between BS/^h, Tab/^i, etc. */
 	t->complex_input = true;
 
@@ -1049,10 +1285,6 @@ static void hook_quit(const char *str) {
 errr init_gcu(int argc, char **argv) {
 	int i;
 
-	/* Initialize info about terminal capabilities */
-	termtype = getenv("TERM");
-	loaded_terminfo = termtype && tgetent(0, termtype) == 1;
-
 	/* Parse args */
 	for (i = 1; i < argc; i++) {
 		if (prefix(argv[i], "-B")) {
@@ -1242,7 +1474,7 @@ errr init_gcu(int argc, char **argv) {
         --------------------------------|----------------------
         <----Term-3: (COLS-30)x7------->| Term-2: 30x(LINES-27)
 
-        composband -mgcu -- -bottom *x7 -right 30x27,* will layout as
+        angband -mgcu -- -bottom *x7 -right 30x27,* will layout as
 
         Term-0: Map (COLS-30)x(LINES-7) | Term-2: 30x27
                                         |------------------------------
@@ -1267,10 +1499,32 @@ errr init_gcu(int argc, char **argv) {
         {
             if (streq(argv[i], "-spacer"))
             {
+                char *pe, *ystr;
+                long lv;
+
                 i++;
                 if (i >= argc)
                     quit("Missing size specifier for -spacer");
-                sscanf(argv[i], "%dx%d", &spacer_cx, &spacer_cy);
+                lv = strtol(argv[i], &pe, 10);
+                /*
+                 * Also reject INT_MIN and INT_MAX so do not have to check
+                 * errno to detect overflow on platforms where sizeof(int) ==
+                 * sizeof(long).
+                 */
+                if (pe == argv[i] || *pe != 'x' || lv <= INT_MIN ||
+                        lv >= INT_MAX) {
+                    quit_fmt("Invalid specification for -spacer; got %s",
+                        argv[i]);
+                }
+                spacer_cx = (int)lv;
+                ystr = pe + 1;
+                lv = strtol(ystr, &pe, 10);
+                if (pe == ystr || !contains_only_spaces(pe) ||
+                        lv <= INT_MIN || lv >= INT_MAX) {
+                    quit_fmt("Invalid specification for -spacer; got %s",
+                        argv[i]);
+                }
+                spacer_cy = (int)lv;
             }
             else if (streq(argv[i], "-right") || streq(argv[i], "-left"))
             {
