@@ -3,9 +3,16 @@
 /*
  * Copyright (c) 1997 Ben Harrison, and others
  *
- * This software may be copied and distributed for educational, research,
- * and not for profit purposes provided that this copyright and statement
- * are included in all such copies.
+ * This work is free software; you can redistribute it and/or modify it
+ * under the terms of either:
+ *
+ * a) the GNU General Public License as published by the Free Software
+ *    Foundation, version 2, or
+ *
+ * b) the "Angband licence":
+ *    This software may be copied and distributed for educational, research,
+ *    and not for profit purposes provided that this copyright and statement
+ *    are included in all such copies.  Other copyrights may also apply.
  */
 
 /* Purpose: Visual Display Support for "z-term.c", for the IBM */
@@ -37,7 +44,7 @@
  * they modify the underlying "ascii" value of the key.  You must use the
  * new "user pref files" to be able to interact with the keypad and such.
  *
- * The "lib/user/pref-ibm.prf" file contains macro definitions and possible
+ * The "lib/user/pref-ibm.prf" file contains keymap definitions and possible
  * alternative color set definitions.  The "lib/user/font-ibm.prf" contains
  * attr/char mappings for walls and floors and such.
  *
@@ -91,6 +98,9 @@
 
 # include <pc.h>
 # include <osfcn.h>
+# include <libc/unconst.h>
+# include <wchar.h>
+# include <wctype.h>
 
 # else /* __DJGPP__ > 1 */
 #  ifdef __DJGPP__
@@ -106,12 +116,12 @@
 # include <conio.h>
 
 /*
- * Hack -- write directly to video card
+ * Write directly to video card
  */
 extern int directvideo = 1;
 
 /*
- * Hack -- no virtual screen
+ * No virtual screen
  */
 # undef USE_VIRTUAL
 
@@ -177,9 +187,9 @@ static int cols = 80;
  * Physical Screen
  */
 #ifdef USE_286
-# define PhysicalScreen ((byte *)MK_FP(0xB800,0x0000))
+# define PhysicalScreen ((uint8_t *)MK_FP(0xB800,0x0000))
 #else
-# define PhysicalScreen ((byte *)(0xB800 << 4))
+# define PhysicalScreen ((uint8_t *)(0xB800 << 4))
 #endif
 
 
@@ -188,7 +198,7 @@ static int cols = 80;
 /*
  * Virtual Screen Contents
  */
-static byte *VirtualScreen;
+static uint8_t *VirtualScreen;
 
 #else
 
@@ -201,7 +211,7 @@ static byte *VirtualScreen;
 
 
 /*
- * Hack -- the cursor "visibility"
+ * The cursor "visibility"
  */
 static int saved_cur_v;
 static int saved_cur_high;
@@ -214,7 +224,7 @@ static int saved_cur_low;
 /*
  * This array is used for "wiping" the screen
  */
-static byte wiper[160];
+static uint8_t wiper[160];
 
 #endif /* USE_CONIO */
 
@@ -228,7 +238,7 @@ static term term_screen_body;
 /*
  * Choose between the "complex" and "simple" color methods
  */
-static byte use_color_complex = FALSE;
+static uint8_t use_color_complex = false;
 
 
 /*
@@ -252,7 +262,7 @@ static long ibm_color_complex[16];
  *
  * Note that many of the choices below suck, but so do crappy monitors.
  */
-static byte ibm_color_simple[16] =
+static uint8_t ibm_color_simple[16] =
 {
 	VID_BLACK,			/* Dark */
 	VID_WHITE,			/* White */
@@ -273,6 +283,37 @@ static byte ibm_color_simple[16] =
 };
 
 
+static uint8_t ext_color_map[32] = {
+	0,
+	1,
+	2,
+	3,
+	4,
+	5,
+	6,
+	7,
+	8,
+	9,
+	10,
+	11,
+	12,
+	13,
+	14,
+	15,
+	10, /* Purple -> Violet */
+	10, /* Violet -> Violet */
+	14, /* Teal -> Light Blue */
+	7,  /* Mud -> Umber */
+	15, /* Light Yellow -> Light Umber */
+	10, /* Magenta -> Violet */
+	14, /* Light Teal -> Light Blue */
+	10, /* Light Violet -> Violet */
+	12, /* Light Pink -> Light Red */
+	7,  /* Mustard -> Yellow */
+	6,  /* Blue Slate -> Blue */
+	14, /* Deep Light Blue -> Light Blue */
+	8   /* Shade -> Light Dark */
+};
 
 /*
  * Activate the "ibm_color_complex" palette information.
@@ -381,6 +422,16 @@ static void activate_color_complex(void)
 
 }
 
+static int get_ext_color(int a)
+{
+	int retval;
+	if (a > -1 && a < 29) {
+		retval = ext_color_map[a];
+	} else {
+		retval = a & 0x0F;
+	}
+	return retval;
+}
 
 /*
  * Note the use of "(x >> 2)" to convert an 8 bit value to a 6 bit value
@@ -395,7 +446,7 @@ static int Term_xtra_ibm_react(void)
 	{
 		long rv, gv, bv, code;
 
-		bool change = FALSE;
+		bool change = false;
 
 		/* Save the default colors */
 		for (i = 0; i < 16; i++)
@@ -412,7 +463,7 @@ static int Term_xtra_ibm_react(void)
 			if (ibm_color_complex[i] != code)
 			{
 				/* Note the change */
-				change = TRUE;
+				change = true;
 
 				/* Apply the desired color */
 				ibm_color_complex[i] = code;
@@ -441,7 +492,7 @@ static int Term_xtra_ibm_react(void)
 
 
 /*
- * Hack -- set the cursor "visibility"
+ * Set the cursor "visibility"
  */
 static void curs_set(int v)
 {
@@ -490,7 +541,7 @@ static void curs_set(int v)
  * the "special" keys, such as keypad keys, function keys, and various keys
  * used in combination with various modifier keys.
  *
- * To simplify this file, we use Angband's "macro processing" ability, in
+ * To simplify this file, we use Angband's "keymap" ability, in
  * combination with a specialized "pref-ibm.prf" file, to handle most of the
  * "special" keys, instead of attempting to fully analyze them here.  This
  * file only has to determine when a "special" key has been pressed, and
@@ -503,7 +554,7 @@ static void curs_set(int v)
  * ("C" for control, "S" for shift, "A" for alt, or any ordered combination),
  * and "SS" encodes the keypress (as the two "digit" hexidecimal encoding of
  * the scan code of the key that was pressed), and the "^_" and "x" and "\r"
- * delimit the encoding for recognition by the macro processing code.
+ * delimit the encoding for recognition by the keymap code.
  *
  * Some important facts about scan codes follow.  All "normal" keys use
  * scan codes from 1-58.  The "function" keys use 59-68 (and 133-134).
@@ -531,24 +582,27 @@ static void curs_set(int v)
  * Shft:  4c35   4d36   4737   4838   4939   5230   532e  (e00d)
  * Ctrl: (8f00)  7400   7700  (8d00)  8400  (9200) (9300) (e00a)
  *
- * See "pref-ibm.prf" for the "standard" macros for various keys.
+ * See "pref-ibm.prf" for the "standard" keymaps for various keys.
  *
  * Certain "bizarre" keypad keys (such as "enter") return a "scan code"
  * of "0xE0", and a "usable" ascii value.  These keys should be treated
  * like the normal keys, see below.  XXX XXX XXX Note that these "special"
  * keys could be prefixed with an optional "ctrl-^" which would allow them
- * to be used in macros without hurting their use in normal situations.
+ * to be used in keymaps without hurting their use in normal situations.
  */
 static errr Term_xtra_ibm_event(int v)
 {
 	int i, k, s;
+	int ch = 0;
 
-	bool mc = FALSE;
-	bool ms = FALSE;
-	bool ma = FALSE;
+	bool mc = false;
+	bool ms = false;
+	bool ma = false;
+	bool kp = false;
 
+	uint8_t mods = 0;
 
-	/* Hack -- Check for a keypress */
+	/* Check for a keypress */
 	if (!v && !bioskey(1)) return (1);
 
 	/* Wait for a keypress */
@@ -567,36 +621,146 @@ static errr Term_xtra_ibm_event(int v)
 	if ((s <= 58) || (s == 0xE0))
 	{
 		/* Enqueue it */
-		if (k) Term_keypress(k);
+		if (k) Term_keypress(k, 0);
 
 		/* Success */
 		return (0);
 	}
 
 	/* Extract the modifier flags */
-	if (i & (1 << K_CTRL)) mc = TRUE;
-	if (i & (1 << K_LSHIFT)) ms = TRUE;
-	if (i & (1 << K_RSHIFT)) ms = TRUE;
-	if (i & (1 << K_ALT)) ma = TRUE;
+	if (i & (1 << K_CTRL)) mc = true;
+	if (i & (1 << K_LSHIFT)) ms = true;
+	if (i & (1 << K_RSHIFT)) ms = true;
+	if (i & (1 << K_ALT)) ma = true;
 
+	switch (s) {
+	case 0x56:
+		/* Some foreign keyboards have a special key on the
+		 * keyboard, which is used to generate the "<", ">", and
+                 * "|" keys (alone, shifted, alt-ed).
+		 */
+		if (ms) {
+			ch = '>';
+		} else if (ma) {
+			ch = '|';
+		} else {
+			ch = '<';
+		}
+		break;
+	case 0x35: kp = true; ch = '/'; break;
+	case 0x37: kp = true; ch = '*'; break;
+	case 0x47: kp = true; ch = '7'; break;
+	case 0x48: kp = true; ch = '8'; break;
+	case 0x49: kp = true; ch = '9'; break;
+	case 0x4A: kp = true; ch = '-'; break;
+	case 0x4B: kp = true; ch = '4'; break;
+	case 0x4C: kp = true; ch = '5'; break;
+	case 0x4D: kp = true; ch = '6'; break;
+	case 0x4E: kp = true; ch = '+'; break;
+	case 0x4F: kp = true; ch = '1'; break;
+	case 0x50: kp = true; ch = '2'; break;
+	case 0x51: kp = true; ch = '3'; break;
+	case 0x52: kp = true; ch = '0'; break;
+	case 0x53: kp = true; ch = '.'; break;
+	case 0x95:
+		if (mc) {
+			kp = true;
+			ch = '/';
+		}
+		break;
+	case 0x96:
+		if (mc) {
+			kp = true;
+			ch = '*';
+		}
+		break;
+	case 0x77:
+		if (mc) {
+			kp = true;
+			ch = '7';
+		}
+		break;
+	case 0x8D:
+		if (mc) {
+			kp = true;
+			ch = '8';
+		}
+		break;
+	case 0x84:
+		if (mc) {
+			kp = true;
+			ch = '9';
+		}
+		break;
+	case 0x8E:
+		if (mc) {
+			kp = true;
+			ch = '-';
+		}
+		break;
+	case 0x73:
+		if (mc) {
+			kp = true;
+			ch = '4';
+		}
+		break;
+	case 0x8F:
+		if (mc) {
+			kp = true;
+			ch = '5';
+		}
+		break;
+	case 0x74:
+		if (mc) {
+			kp = true;
+			ch = '6';
+		}
+		break;
+	case 0x90:
+		if (mc) {
+			kp = true;
+			ch = '+';
+		}
+		break;
+	case 0x75:
+		if (mc) {
+			kp = true;
+			ch = '1';
+		}
+		break;
+	case 0x91:
+		if (mc) {
+			kp = true;
+			ch = '2';
+		}
+		break;
+	case 0x76:
+		if (mc) {
+			kp = true;
+			ch = '3';
+		}
+		break;
+	case 0x92:
+		if (mc) {
+			kp = true;
+			ch = '0';
+		}
+		break;
+	case 0x93:
+		if (mc) {
+			kp = true;
+			ch = '.';
+		}
+		break;
+	default: break;
+	}
 
-	/* Begin a "macro trigger" */
-	Term_keypress(31);
-
-	/* Hack -- Send the modifiers */
-	if (mc) Term_keypress('C');
-	if (ms) Term_keypress('S');
-	if (ma) Term_keypress('A');
-
-	/* Introduce the hexidecimal scan code */
-	Term_keypress('x');
-
-	/* Encode the hexidecimal scan code */
-	Term_keypress(hexsym[s/16]);
-	Term_keypress(hexsym[s%16]);
-
-	/* End the "macro trigger" */
-	Term_keypress(13);
+	if (ch) {
+		if (kp) mods |= KC_MOD_KEYPAD;
+		if (ma) mods |= KC_MOD_ALT;
+		if (ms) mods |= KC_MOD_SHIFT;
+		Term_keypress(ch, mods);
+	}
 
 	/* Success */
 	return (0);
@@ -710,7 +874,7 @@ static errr Term_xtra_ibm(int n, int v)
 		case TERM_XTRA_FLUSH:
 		{
 			/* Strip events */
-			while (!Term_xtra_ibm_event(FALSE)) /* loop */;
+			while (!Term_xtra_ibm_event(false)) /* loop */;
 
 			/* Success */
 			return (0);
@@ -815,25 +979,24 @@ static errr Term_wipe_ibm(int x, int y, int n)
  *
  * The string "cp" has length "n" and is NOT null-terminated.
  */
-static errr Term_text_ibm(int x, int y, int n, byte a, const char *cp)
+static errr Term_text_ibm(int x, int y, int n, int a, const wchar_t *cp)
 {
 	register int i;
-	register byte attr;
-	register byte *dest;
-
+	register uint8_t attr;
+	register uint8_t *dest;
 
 	/* Handle "complex" color */
 	if (use_color_complex)
 	{
 		/* Extract a color index */
-		attr = (a & 0x0F);
+		attr = (uint8_t)(get_ext_color(a));
 	}
 
 	/* Handle "simple" color */
 	else
 	{
 		/* Extract a color value */
-		attr = ibm_color_simple[a & 0x0F];
+		attr = ibm_color_simple[get_ext_color(a)];
 	}
 
 #ifdef USE_CONIO
@@ -845,7 +1008,11 @@ static errr Term_text_ibm(int x, int y, int n, byte a, const char *cp)
 	textattr(attr);
 
 	/* Dump the text */
-	for (i = 0; i < n; i++) putch(cp[i]);
+	for (i = 0; i < n; i++) {
+		if (cp[i] < 0x100) {
+			putch(cp[i]);
+		}
+	}
 
 #else /* USE_CONIO */
 
@@ -872,12 +1039,12 @@ static errr Term_text_ibm(int x, int y, int n, byte a, const char *cp)
  *
  * The given parameters are "valid".
  */
-static errr Term_pict_ibm(int x, int y, int n, const byte *ap, const char *cp,
-                          const byte *tap, const char *tcp)
+static errr Term_pict_ibm(int x, int y, int n, const int *ap,
+			  const wchar_t *cp, const int *tap, const wchar_t *tcp)
 {
 	register int i;
-	register byte attr;
-	register byte *dest;
+	register uint8_t attr;
+	register uint8_t *dest;
 
 	/* Unused parameters */
 	(void)tap;
@@ -895,21 +1062,23 @@ static errr Term_pict_ibm(int x, int y, int n, const byte *ap, const char *cp,
 		if (use_color_complex)
 		{
 			/* Extract a color index */
-			attr = (ap[i] & 0x0F);
+			attr = (uint8_t)(get_ext_color(ap[i]));
 		}
 
 		/* Handle "simple" color */
 		else
 		{
 			/* Extract a color value */
-			attr = ibm_color_simple[ap[i] & 0x0F];
+			attr = ibm_color_simple[get_ext_color(ap[i])];
 		}
 
 		/* Set the attribute */
 		textattr(attr);
 
 		/* Dump the char */
-		putch(cp[i]);
+		if (cp[i] < 0x100) {
+			putch(cp[i]);
+		}
 	}
 
 #else /* USE_CONIO */
@@ -924,14 +1093,14 @@ static errr Term_pict_ibm(int x, int y, int n, const byte *ap, const char *cp,
 		if (use_color_complex)
 		{
 			/* Extract a color index */
-			attr = (ap[i] & 0x0F);
+			attr = (uint8_t)(get_ext_color(ap[i]));
 		}
 
 		/* Handle "simple" color */
 		else
 		{
 			/* Extract a color value */
-			attr = ibm_color_simple[ap[i] & 0x0F];
+			attr = ibm_color_simple[get_ext_color(ap[i])];
 		}
 
 		/* Apply */
@@ -997,7 +1166,41 @@ static void Term_nuke_ibm(term *t)
 	curs_set(1);
 }
 
+int Term_iswprint_ibm(wint_t wc)
+{
+	int retval;
+	retval = (wc < 0x100);
+	return retval;
+}
 
+wchar_t *Term_wcschr_ibm(const wchar_t *s, wchar_t c) {
+	wchar_t cc = c;
+
+	while (*s) {
+		if (*s == cc) {
+			return unconst(s, wchar_t *);
+		}
+		s++;
+	}
+        
+	if (cc == 0) {
+		return unconst(s, wchar_t *);
+	}
+
+	return 0;
+}
+
+size_t Term_wcslen_ibm(const wchar_t *s) {
+	const wchar_t *p;
+
+	if (s == 0) {
+		return 0;
+	}
+
+	for (p = s; *p; ++p);
+
+	return p - s;
+}
 
 #ifdef USE_GRAPHICS
 
@@ -1196,7 +1399,7 @@ const char help_ibm[] = "IBM Visual Display Support";
 /*
  * Initialize the IBM "visual module"
  *
- * Hack -- we assume that "blank space" should be "white space"
+ * We assume that "blank space" should be "white space"
  * (and not "black space" which might make more sense).
  *
  * Note the use of "((x << 2) | (x >> 4))" to "expand" a 6 bit value
@@ -1283,7 +1486,7 @@ errr init_ibm(int argc, char **argv)
 	if (mode == 0x13)
 	{
 		/* Remember the mode */
-		use_color_complex = TRUE;
+		use_color_complex = true;
 
 		/* Instantiate the color set */
 		activate_color_complex();
@@ -1339,7 +1542,7 @@ errr init_ibm(int argc, char **argv)
 		wiper[2*i] = ' ';
 
 		/* Black */
-		wiper[2*i+1] = TERM_WHITE;
+		wiper[2*i+1] = COLOUR_WHITE;
 	}
 
 #endif /* USE_CONIO */
@@ -1348,7 +1551,7 @@ errr init_ibm(int argc, char **argv)
 #ifdef USE_VIRTUAL
 
 	/* Make the virtual screen */
-	C_MAKE(VirtualScreen, rows * cols * 2, byte);
+	VirtualScreen = mem_zalloc(rows * cols * 2 * sizeof(uint8_t));
 
 #endif /* USE_VIRTUAL */
 
@@ -1381,14 +1584,10 @@ errr init_ibm(int argc, char **argv)
 #else /* USE_CONIO */
 
 	/* Always use "Term_pict()" */
-	t->always_pict = TRUE;
+	t->always_pict = true;
 
 #endif /* USE_CONIO */
 
-	/* Use "white space" to erase */
-	t->attr_blank = TERM_WHITE;
-	t->char_blank = ' ';
-
 	/* Prepare the init/nuke hooks */
 	t->init_hook = Term_init_ibm;
 	t->nuke_hook = Term_nuke_ibm;
@@ -1399,6 +1598,9 @@ errr init_ibm(int argc, char **argv)
 	t->wipe_hook = Term_wipe_ibm;
 	t->text_hook = Term_text_ibm;
 	t->pict_hook = Term_pict_ibm;
+        text_iswprint_hook = Term_iswprint_ibm;
+        text_wcschr_hook = Term_wcschr_ibm;
+        text_wcslen_hook = Term_wcslen_ibm;
 
 	/* Save it */
 	term_screen = t;
