@@ -46,7 +46,6 @@
 #include "player-util.h"
 #include "project.h"
 #include "trap.h"
-#include "z-set.h"
 
 /**
  * ------------------------------------------------------------------------
@@ -217,7 +216,7 @@ static void path_analyse(struct chunk *c, struct loc grid)
 	}
 
 	/* Plot the path. */
-	path_n = project_path(cave, path_g, z_info->max_range, player->grid,
+	path_n = project_path(c, path_g, z_info->max_range, player->grid,
 		grid, PROJECT_NONE);
 
 	/* Project along the path */
@@ -419,7 +418,7 @@ void update_mon(struct monster *mon, struct chunk *c, bool full)
 			if (player->upkeep->health_who == mon)
 				player->upkeep->redraw |= (PR_HEALTH);
 
-			/* Hack -- Count "fresh" sightings */
+			/* Count "fresh" sightings */
 			if (lore->sights < SHRT_MAX)
 				lore->sights++;
 
@@ -467,7 +466,7 @@ void update_mon(struct monster *mon, struct chunk *c, bool full)
 			mflag_off(mon->mflag, MFLAG_VIEW);
 
 			/* Disturb on disappearance */
-			if (OPT(player, disturb_near) && !monster_is_mimicking(mon))
+			if (OPT(player, disturb_near) && !monster_is_camouflaged(mon))
 				disturb(player);
 
 			/* Re-draw monster list window */
@@ -584,16 +583,16 @@ void monster_swap(struct loc grid1, struct loc grid2)
 		mon = cave_monster(cave, m1);
 
 		/* Update monster */
-		if (monster_is_mimicking(mon)) {
+		if (monster_is_camouflaged(mon)) {
 			/*
-			 * Become aware if the player can see the mimic before
-			 * or after the swap.
+			 * Become aware if the player can see the grid with
+			 * the camouflaged monster before or after the swap.
 			 */
 			if (monster_is_in_view(mon) ||
 				(m2 >= 0 && los(cave, pgrid, grid2)) ||
 				(m2 < 0 && los(cave, grid1, grid2))) {
 				become_aware(cave, mon);
-			} else {
+			} else if (monster_is_mimicking(mon)) {
 				move_mimicked_object(cave, mon, grid1, grid2);
 				player->upkeep->redraw |= (PR_ITEMLIST);
 			}
@@ -631,16 +630,16 @@ void monster_swap(struct loc grid1, struct loc grid2)
 		mon = cave_monster(cave, m2);
 
 		/* Update monster */
-		if (monster_is_mimicking(mon)) {
+		if (monster_is_camouflaged(mon)) {
 			/*
-			 * Become aware if the player can see the mimic before
-			 * or after the swap.
+			 * Become aware if the player can see the grid with
+			 * the camouflaged monster before or after the swap.
 			 */
 			if (monster_is_in_view(mon) ||
 				(m1 >= 0 && los(cave, pgrid, grid1)) ||
 				(m1 < 0 && los(cave, grid2, grid1))) {
 				become_aware(cave, mon);
-			} else {
+			} else if (monster_is_mimicking(mon)) {
 				move_mimicked_object(cave, mon, grid2, grid1);
 				player->upkeep->redraw |= (PR_ITEMLIST);
 			}
@@ -903,28 +902,28 @@ bool find_any_nearby_injured_kin(struct chunk *c, const struct monster *mon)
  * Choose one injured monster of the same base in LOS of the provided monster.
  *
  * Scan MAX_KIN_RADIUS grids around the monster to find potential grids,
- * make a list of kin, and choose a random one.
+ * using reservoir sampling with k = 1 to find a random one.
  */
 struct monster *choose_nearby_injured_kin(struct chunk *c,
-										  const struct monster *mon)
+                                          const struct monster *mon)
 {
-	struct set *set = set_new();
 	struct loc grid;
+	int nseen = 0;
+	struct monster *found = NULL;
 
 	for (grid.y = mon->grid.y - MAX_KIN_RADIUS;
 		 grid.y <= mon->grid.y + MAX_KIN_RADIUS; grid.y++) {
 		for (grid.x = mon->grid.x - MAX_KIN_RADIUS;
 			 grid.x <= mon->grid.x + MAX_KIN_RADIUS; grid.x++) {
 			struct monster *kin = get_injured_kin(c, mon, grid);
-			if (kin != NULL) {
-				set_add(set, kin);
+			if (kin) {
+				nseen++;
+				if (!randint0(nseen))
+					found = kin;
 			}
 		}
 	}
 
-	struct monster *found = set_choose(set);
-	set_free(set);
-
 	return found;
 }
 
@@ -1032,7 +1031,7 @@ static void player_kill_monster(struct monster *mon, struct player *p,
 	}
 
 	/* Play a special sound if the monster was unique */
-	if (rf_has(mon->race->flags, RF_UNIQUE)) {
+	if (monster_is_unique(mon)) {
 		if (mon->race->base == lookup_monster_base("Morgoth"))
 			soundfx = MSG_KILL_KING;
 		else
@@ -1085,7 +1084,7 @@ static void player_kill_monster(struct monster *mon, struct player *p,
 	}
 
 	/* When the player kills a Unique, it stays dead */
-	if (rf_has(mon->race->flags, RF_UNIQUE)) {
+	if (monster_is_unique(mon)) {
 		char unique_name[80];
 		assert(mon->original_race == NULL);
 		mon->race->max_num = 0;
@@ -1110,7 +1109,7 @@ static void player_kill_monster(struct monster *mon, struct player *p,
 
 	/* Bloodlust bonus */
 	if (p->timed[TMD_BLOODLUST]) {
-		player_inc_timed(p, TMD_BLOODLUST, 10, false, false);
+		player_inc_timed(p, TMD_BLOODLUST, 10, false, false, true);
 		player_over_exert(p, PY_EXERT_CONF, 5, 3);
 		player_over_exert(p, PY_EXERT_HALLU, 5, 10);
 	}
@@ -1129,7 +1128,7 @@ static void player_kill_monster(struct monster *mon, struct player *p,
 	}
 
 	/* Delete the monster */
-	delete_monster_idx(mon->midx);
+	delete_monster_idx(cave, mon->midx);
 }
 
 /**
@@ -1185,9 +1184,11 @@ static bool monster_scared_by_damage(struct monster *mon, int dam)
  * This is a helper for melee handlers. It is very similar to mon_take_hit(),
  * but eliminates the player-oriented stuff of that function.
  *
- * \param context is the project_m context.
- * \param hurt_msg is the message if the monster is hurt (if any).
- * \return true if the monster died, false if it is still alive.
+ * \param dam is the amount of damage to inflict
+ * \param t_mon is the monster to damage
+ * \param hurt_msg is the message, if any, to use when the monster is hurt
+ * \param die_msg is the message, if any to use when the monster dies
+ * \return true if the monster died, false if it is still alive
  */
 bool mon_take_nonplayer_hit(int dam, struct monster *t_mon,
 							enum mon_messages hurt_msg,
@@ -1196,8 +1197,7 @@ bool mon_take_nonplayer_hit(int dam, struct monster *t_mon,
 	assert(t_mon);
 
 	/* "Unique" or arena monsters can only be "killed" by the player */
-	if (rf_has(t_mon->race->flags, RF_UNIQUE)
-			|| player->upkeep->arena_level) {
+	if (monster_is_unique(t_mon) || player->upkeep->arena_level) {
 		/* Reduce monster hp to zero, but don't kill it. */
 		if (dam > t_mon->hp) dam = t_mon->hp;
 	}
@@ -1214,6 +1214,11 @@ bool mon_take_nonplayer_hit(int dam, struct monster *t_mon,
 
 	/* Dead or damaged monster */
 	if (t_mon->hp < 0) {
+		/* Shapechanged monsters revert on death */
+		if (t_mon->original_race) {
+			monster_revert_shape(t_mon);
+		}
+
 		/* Death message */
 		add_monster_message(t_mon, die_msg, false);
 
@@ -1221,9 +1226,9 @@ bool mon_take_nonplayer_hit(int dam, struct monster *t_mon,
 		monster_death(t_mon, player, false);
 
 		/* Delete the monster */
-		delete_monster_idx(t_mon->midx);
+		delete_monster_idx(cave, t_mon->midx);
 		return true;
-	} else if (!monster_is_mimicking(t_mon)) {
+	} else if (!monster_is_camouflaged(t_mon)) {
 		/* Give detailed messages if visible */
 		if (hurt_msg != MON_MSG_NONE) {
 			add_monster_message(t_mon, hurt_msg, false);
@@ -1243,7 +1248,7 @@ bool mon_take_nonplayer_hit(int dam, struct monster *t_mon,
 /**
  * Decreases a monster's hit points by `dam` and handle monster death.
  *
- * Hack -- we "delay" fear messages by passing around a "fear" flag.
+ * We "delay" fear messages by passing around a "fear" flag.
  *
  * We announce monster death (using an optional "death message" (`note`)
  * if given, and a otherwise a generic killed/destroyed message).
@@ -1434,7 +1439,7 @@ void steal_monster_item(struct monster *mon, int midx)
 
 	if (midx < 0) {
 		/* Base monster protection and player stealing skill */
-		bool unique = rf_has(mon->race->flags, RF_UNIQUE);
+		bool unique = monster_is_unique(mon);
 		int guard = (mon->race->level * (unique ? 4 : 3)) / 4 +
 			mon->mspeed - player->state.speed;
 		int steal_skill = player->state.skills[SKILL_STEALTH] +
@@ -1462,7 +1467,7 @@ void steal_monster_item(struct monster *mon, int midx)
 
 		/* Monster base reaction, plus allowance for item weight */
 		monster_reaction = guard / 2 + randint1(MAX(guard, 1));
-		monster_reaction += obj->weight / 20;
+		monster_reaction += (obj->number * object_weight_one(obj)) / 20;
 
 		/* Try and steal */
 		if (monster_reaction < steal_skill) {
@@ -1529,7 +1534,8 @@ void steal_monster_item(struct monster *mon, int midx)
 			msg("You vanish into the shadows!");
 			effect_simple(EF_TELEPORT, source_player(), near, 0, 0, 0, 0, 0,
 						  NULL);
-			(void) player_clear_timed(player, TMD_ATT_RUN, false);
+			(void) player_clear_timed(player, TMD_ATT_RUN, false,
+				false);
 		}
 	} else {
 		/* Get the thief details */
