@@ -19,19 +19,22 @@
 #include "angband.h"
 #include "cave.h"
 #include "cmd-core.h"
-#include "cmds.h"
 #include "game-input.h"
 #include "game-world.h"
 #include "generate.h"
 #include "init.h"
+#include "mon-predicate.h"
 #include "obj-chest.h"
 #include "obj-gear.h"
+#include "obj-ignore.h"
 #include "obj-knowledge.h"
 #include "obj-pile.h"
 #include "obj-tval.h"
 #include "obj-util.h"
+#include "player-attack.h"
 #include "player-calcs.h"
 #include "player-history.h"
+#include "player-quest.h"
 #include "player-spell.h"
 #include "player-timed.h"
 #include "player-util.h"
@@ -40,6 +43,7 @@
 #include "store.h"
 #include "target.h"
 #include "trap.h"
+#include "ui-input.h"
 
 /**
  * Increment to the next or decrement to the preceeding level
@@ -47,7 +51,7 @@
    Keep in mind to check all intermediate level for unskippable
    quests
 */
-int dungeon_get_next_level(int dlev, int added)
+int dungeon_get_next_level(struct player *p, int dlev, int added)
 {
 	int target_level, i;
 
@@ -63,7 +67,7 @@ int dungeon_get_next_level(int dlev, int added)
 
 	/* Check intermediate levels for quests */
 	for (i = dlev; i <= target_level; i++) {
-		if (is_quest(i)) return i;
+		if (is_quest(p, i)) return i;
 	}
 
 	return target_level;
@@ -77,8 +81,10 @@ void player_set_recall_depth(struct player *p)
 	/* Account for forced descent */
 	if (OPT(p, birth_force_descend)) {
 		/* Force descent to a lower level if allowed */
-		if ((p->max_depth < z_info->max_depth - 1) && !is_quest(p->max_depth)) {
-			p->recall_depth = dungeon_get_next_level(p->max_depth, 1);
+		if (p->max_depth < z_info->max_depth - 1
+				&& !is_quest(p, p->max_depth)) {
+			p->recall_depth = dungeon_get_next_level(p,
+				p->max_depth, 1);
 		}
 	}
 
@@ -96,8 +102,16 @@ bool player_get_recall_depth(struct player *p)
 	bool level_ok = false;
 	int new = 0;
 
+	/*
+	 * No choice when have not entered the dungeon or descent is forced,
+	 * so do not prompt.
+	 */
+	if (p->max_depth <= 0 || OPT(p, birth_force_descend)) {
+		return true;
+	}
 	while (!level_ok) {
-		char *prompt = "Which level do you wish to return to (0 to cancel)? ";
+		const char *prompt =
+			"Which level do you wish to return to (0 to cancel)? ";
 		int i;
 
 		/* Choose the level */
@@ -142,10 +156,41 @@ void dungeon_change_level(struct player *p, int dlev)
 }
 
 
+/**
+ * Returns what an incoming damage amount would be after applying a player's
+ * damage reduction.
+ *
+ * \param p is the player of interest.
+ * \param dam is the incoming damaage amount.
+ * \return the damage after the player's damage reduction, if any.
+ */
+int player_apply_damage_reduction(struct player *p, int dam)
+{
+	/* Mega-Hack -- Apply "invulnerability" */
+	if (p->timed[TMD_INVULN] && (dam < 9000)) return 0;
+
+	dam -= p->state.dam_red;
+	if (dam > 0 && p->state.perc_dam_red) {
+		dam -= (dam * p->state.perc_dam_red) / 100 ;
+	}
+
+	return (dam < 0) ? 0 : dam;
+}
+
+
 /**
  * Decreases players hit points and sets death flag if necessary
  *
- * Hack -- this function allows the user to save (or quit) the game
+ * \param p is the player of interest.
+ * \param dam is the amount of damage to apply.  If dam is less than
+ * or equal to zero, nothing will be done.  The amount of damage should have
+ * been processed with player_apply_damage_reduction(); that is not done
+ * internally here so the caller can display messages that include the amount of
+ * damage.
+ * \param kb_str is the null-terminated string describing the cause of the
+ * damage.
+ *
+ * This function allows the user to save (or quit) the game
  * when he dies, since the "You die." message is shown before setting
  * the player to "dead".
  */
@@ -156,17 +201,7 @@ void take_hit(struct player *p, int dam, const char *kb_str)
 	int warning = (p->mhp * p->opts.hitpoint_warn / 10);
 
 	/* Paranoia */
-	if (p->is_dead) return;
-
-	/* Mega-Hack -- Apply "invulnerability" */
-	if (p->timed[TMD_INVULN] && (dam < 9000)) return;
-
-	/* Apply damage reduction */
-	dam -= p->state.dam_red;
-	if (p->state.perc_dam_red) {
-		dam -= (dam * p->state.perc_dam_red) / 100 ;
-	}
-	if (dam <= 0) return;
+	if (p->is_dead || dam <= 0) return;
 
 	/* Disturb */
 	disturb(p);
@@ -178,10 +213,11 @@ void take_hit(struct player *p, int dam, const char *kb_str)
 	 * Unenviable task of separating what should and should not cause rage
 	 * If we eliminate the most exploitable cases it should be fine.
 	 * All traps and lava currently give mana, which could be exploited  */
-	if (player_has(p, PF_COMBAT_REGEN)  && strcmp(kb_str, "poison")
-		&& strcmp(kb_str, "a fatal wound") && strcmp(kb_str, "starvation")) {
+	if (player_has(p, PF_COMBAT_REGEN)  && !streq(kb_str, "poison")
+		&& !streq(kb_str, "a fatal wound") && !streq(kb_str, "starvation")) {
 		/* lose X% of hitpoints get X% of spell points */
-		s32b sp_gain = (MAX((s32b)p->msp, 10) << 16) / (s32b)p->mhp * dam;
+		int32_t sp_gain = (((int32_t)MAX(p->msp, 10)) * 65536)
+			/ (int32_t)p->mhp * dam;
 		player_adjust_mana_precise(p, sp_gain);
 	}
 
@@ -199,32 +235,39 @@ void take_hit(struct player *p, int dam, const char *kb_str)
 				msg("So great was his prowess and skill in warfare, the Elves said: ");
 				msg("'The Mormegil cannot be slain, save by mischance.'");
 			}
-		} else if ((p->wizard || OPT(p, cheat_live)) && !get_check("Die? ")) {
-			event_signal(EVENT_CHEAT_DEATH);
 		} else {
-			/* Hack -- Note death */
-			msgt(MSG_DEATH, "You die.");
-			event_signal(EVENT_MESSAGE_FLUSH);
-
-			/* Note cause of death */
+			/*
+			 * Note cause of death.  Do it here so EVENT_CHEAT_DEATH
+			 * handlers or things looking for the "Die? " prompt
+			 * (the borg, for instance), have access to it.
+			 */
 			my_strcpy(p->died_from, kb_str, sizeof(p->died_from));
 
-			/* No longer a winner */
-			p->total_winner = false;
+			if ((p->wizard || OPT(p, cheat_live))
+					&& !get_check("Die? ")) {
+				event_signal(EVENT_CHEAT_DEATH);
+			} else {
+				/* Note death */
+				msgt(MSG_DEATH, "You die.");
+				event_signal(EVENT_MESSAGE_FLUSH);
 
-			/* Note death */
-			p->is_dead = true;
+				/* No longer a winner */
+				p->total_winner = false;
 
-			/* Dead */
-			return;
+				/* Note death */
+				p->is_dead = true;
+
+				/* Dead */
+				return;
+			}
 		}
 	}
 
 	/* Hitpoint warning */
 	if (p->chp < warning) {
-		/* Hack -- bell on first notice */
+		/* Bell on first notice */
 		if (old_chp > warning)
-			bell("Low hitpoint warning!");
+			bell();
 
 		/* Message */
 		msgt(MSG_HITPOINT_WARN, "*** LOW HITPOINT WARNING! ***");
@@ -237,14 +280,14 @@ void take_hit(struct player *p, int dam, const char *kb_str)
  */
 void death_knowledge(struct player *p)
 {
-	struct store *home = &stores[STORE_HOME];
+	struct store *home = &stores[f_info[FEAT_HOME].shopnum - 1];
 	struct object *obj;
 	time_t death_time = (time_t)0;
 
 	/* Retire in the town in a good state */
 	if (p->total_winner) {
 		p->depth = 0;
-		my_strcpy(p->died_from, "Ripe Old Age", sizeof(p->died_from));
+		my_strcpy(p->died_from, WINNING_HOW, sizeof(p->died_from));
 		p->exp = p->max_exp;
 		p->lev = p->max_lev;
 		p->au += 10000000L;
@@ -252,13 +295,13 @@ void death_knowledge(struct player *p)
 
 	player_learn_all_runes(p);
 	for (obj = p->gear; obj; obj = obj->next) {
-		object_flavor_aware(obj);
+		object_flavor_aware(p, obj);
 		obj->known->effect = obj->effect;
 		obj->known->activation = obj->activation;
 	}
 
 	for (obj = home->stock; obj; obj = obj->next) {
-		object_flavor_aware(obj);
+		object_flavor_aware(p, obj);
 		obj->known->effect = obj->effect;
 		obj->known->activation = obj->activation;
 	}
@@ -267,13 +310,23 @@ void death_knowledge(struct player *p)
 
 	/* Get time of death */
 	(void)time(&death_time);
-	enter_score(&death_time);
+	enter_score(p, &death_time);
 
-	/* Hack -- Recalculate bonuses */
+	/* Recalculate bonuses */
 	p->upkeep->update |= (PU_BONUS);
 	handle_stuff(p);
 }
 
+/**
+ * Energy per move, taking extra moves into account
+ */
+int energy_per_move(struct player *p)
+{
+	int num = p->state.num_moves;
+	int energy = z_info->move_energy;
+	return (energy * (1 + ABS(num) - num)) / (1 + ABS(num));
+}
+
 /**
  * Modify a stat value by a "modifier", return new value
  *
@@ -283,7 +336,7 @@ void death_knowledge(struct player *p)
  * Stats go down: 18/220, 18/210,..., 18/10, 18, 17, ..., 3
  * Or even: 18/13, 18/03, 18, 17, ..., 3
  */
-s16b modify_stat_value(int value, int amount)
+int16_t modify_stat_value(int value, int amount)
 {
 	int i;
 
@@ -303,7 +356,7 @@ s16b modify_stat_value(int value, int amount)
 			/* Ten points at a time */
 			if (value >= 18+10) value -= 10;
 
-			/* Hack -- prevent weirdness */
+			/* Prevent weirdness */
 			else if (value > 18) value = 18;
 
 			/* One point at a time */
@@ -315,12 +368,74 @@ s16b modify_stat_value(int value, int amount)
 	return (value);
 }
 
+/**
+ * Swap player's stats at random, retaining information so they can be
+ * reverted to their original state.
+ */
+void player_scramble_stats(struct player *p)
+{
+	int max1, cur1, max2, cur2, i, j, swap;
+
+	/* Fisher-Yates shuffling algorithm */
+	for (i = STAT_MAX - 1; i > 0; --i) {
+		j = randint0(i);
+
+		max1 = p->stat_max[i];
+		cur1 = p->stat_cur[i];
+		max2 = p->stat_max[j];
+		cur2 = p->stat_cur[j];
+
+		p->stat_max[i] = max2;
+		p->stat_cur[i] = cur2;
+		p->stat_max[j] = max1;
+		p->stat_cur[j] = cur1;
+
+		/* Record what we did */
+		swap = p->stat_map[i];
+		assert(swap >= 0 && swap < STAT_MAX);
+		p->stat_map[i] = p->stat_map[j];
+		assert(p->stat_map[i] >= 0 && p->stat_map[i] < STAT_MAX);
+		p->stat_map[j] = swap;
+	}
+
+	/* Mark what else needs to be updated */
+	p->upkeep->update |= (PU_BONUS);
+}
+
+/**
+ * Revert all prior swaps to the player's stats.  Has no effect if the
+ * stats have not been swapped.
+ */
+void player_fix_scramble(struct player *p)
+{
+	/* Figure out what stats should be */
+	int new_cur[STAT_MAX];
+	int new_max[STAT_MAX];
+	int i;
+
+	for (i = 0; i < STAT_MAX; ++i) {
+		assert(p->stat_map[i] >= 0 && p->stat_map[i] < STAT_MAX);
+		new_cur[p->stat_map[i]] = p->stat_cur[i];
+		new_max[p->stat_map[i]] = p->stat_max[i];
+	}
+
+	/* Apply new stats and reset stat_map */
+	for (i = 0; i < STAT_MAX; ++i) {
+		p->stat_cur[i] = new_cur[i];
+		p->stat_max[i] = new_max[i];
+		p->stat_map[i] = i;
+	}
+
+	/* Mark what else needs to be updated */
+	p->upkeep->update |= (PU_BONUS);
+}
+
 /**
  * Regenerate one turn's worth of hit points
  */
 void player_regen_hp(struct player *p)
 {
-	s32b hp_gain;
+	int32_t hp_gain;
 	int percent = 0;/* max 32k -> 50% of mhp; more accurately "pertwobytes" */
 	int fed_pct, old_chp = p->chp;
 
@@ -345,7 +460,7 @@ void player_regen_hp(struct player *p)
 		percent *= 2;
 
 	/* Some things slow it down */
-	if (player_of_has(p, OF_IMPAIR_HP) || player_has(p, PF_COMBAT_REGEN))
+	if (player_of_has(p, OF_IMPAIR_HP))
 		percent /= 2;
 
 	/* Various things interfere with physical healing */
@@ -355,7 +470,7 @@ void player_regen_hp(struct player *p)
 	if (p->timed[TMD_CUT]) percent = 0;
 
 	/* Extract the new hitpoints */
-	hp_gain = (s32b)(p->mhp * percent) + PY_REGEN_HPBASE;
+	hp_gain = (int32_t)(p->mhp * percent) + PY_REGEN_HPBASE;
 	player_adjust_hp_precise(p, hp_gain);
 
 	/* Notice changes */
@@ -371,7 +486,7 @@ void player_regen_hp(struct player *p)
  */
 void player_regen_mana(struct player *p)
 {
-	s32b sp_gain;
+	int32_t sp_gain;
 	int percent, old_csp = p->csp;
 
 	/* Save the old spell points */
@@ -396,14 +511,14 @@ void player_regen_mana(struct player *p)
 	}
 
 	/* Regenerate mana */
-	sp_gain = (s32b)(p->msp * percent);
+	sp_gain = (int32_t)(p->msp * percent);
 	if (percent >= 0)
 		sp_gain += PY_REGEN_MNBASE;
 	sp_gain = player_adjust_mana_precise(p, sp_gain);
 
 	/* SP degen heals BGs at double efficiency vs casting */
 	if (sp_gain < 0  && player_has(p, PF_COMBAT_REGEN)) {
-		convert_mana_to_hp(p, -sp_gain << 2);
+		convert_mana_to_hp(p, -sp_gain * 2);
 	}
 
 	/* Notice changes */
@@ -414,27 +529,42 @@ void player_regen_mana(struct player *p)
 	}
 }
 
-void player_adjust_hp_precise(struct player *p, s32b hp_gain)
+void player_adjust_hp_precise(struct player *p, int32_t hp_gain)
 {
-	s32b new_chp;
-	int num, old_chp = p->chp;
-
-	/* Load it all into 4 byte format*/
-	new_chp = (s32b)((p->chp << 16) + p->chp_frac) + hp_gain;
+	int16_t old_16 = p->chp;
+	/* Load it all into 4 byte format */
+	int32_t old_32 = ((int32_t) old_16) * 65536 + p->chp_frac, new_32;
 
 	/* Check for overflow */
-	/*     {new_chp = LONG_MIN;} DAVIDTODO*/
-	if ((new_chp < 0) && (old_chp > 0) && (hp_gain > 0)) {
-		new_chp = INT32_MAX;
-	} else if ((new_chp > 0) && (old_chp < 0) && (hp_gain < 0)) {
-		new_chp = INT32_MIN;
+	if (hp_gain >= 0) {
+		new_32 = (old_32 < INT32_MAX - hp_gain) ?
+			old_32 + hp_gain : INT32_MAX;
+	} else {
+		new_32 = (old_32 > INT32_MIN - hp_gain) ?
+			old_32 + hp_gain : INT32_MIN;
 	}
 
-	/* Break it back down*/
-	p->chp = (s16b)(new_chp >> 16);   /* div 65536 */
-	p->chp_frac = (u16b)(new_chp & 0xFFFF); /* mod 65536 */
-	/*DAVIDTODO neg new_chp ok? I think so because eg a slightly negative
-	 * new_chp will give -1 for chp and very high chp_frac.*/
+	/* Break it back down */
+	if (new_32 < 0) {
+		/*
+		 * Don't use right bitwise shift on negative values:  whether
+		 * the left bits are zero or one depends on the system.
+		 */
+		int32_t remainder = new_32 % 65536;
+
+		p->chp = (int16_t) (new_32 / 65536);
+		if (remainder) {
+			assert(remainder < 0);
+			p->chp_frac = (uint16_t) (65536 + remainder);
+			assert(p->chp > INT16_MIN);
+			p->chp -= 1;
+		} else {
+			p->chp_frac = 0;
+		}
+	} else {
+		p->chp = (int16_t)(new_32 >> 16);   /* div 65536 */
+		p->chp_frac = (uint16_t)(new_32 & 0xFFFF); /* mod 65536 */
+	}
 
 	/* Fully healed */
 	if (p->chp >= p->mhp) {
@@ -442,11 +572,9 @@ void player_adjust_hp_precise(struct player *p, s32b hp_gain)
 		p->chp_frac = 0;
 	}
 
-	num = p->chp - old_chp;
-	if (num == 0)
-		return;
-
-	p->upkeep->redraw |= (PR_HP);
+	if (p->chp != old_16) {
+		p->upkeep->redraw |= (PR_HP);
+	}
 }
 
 
@@ -454,31 +582,50 @@ void player_adjust_hp_precise(struct player *p, s32b hp_gain)
  * Accept a 4 byte signed int, divide it by 65k, and add
  * to current spell points. p->csp and csp_frac are 2 bytes each.
  */
-s32b player_adjust_mana_precise(struct player *p, s32b sp_gain)
+int32_t player_adjust_mana_precise(struct player *p, int32_t sp_gain)
 {
-	s32b old_csp_long, new_csp_long;
-	int old_csp_short = p->csp;
+	int16_t old_16 = p->csp;
+	/* Load it all into 4 byte format*/
+	int32_t old_32 = ((int32_t) p->csp) * 65536 + p->csp_frac, new_32;
 
 	if (sp_gain == 0) return 0;
 
-	/* Load it all into 4 byte format*/
-	old_csp_long = (s32b)((p->csp << 16) + p->csp_frac);
-	new_csp_long = old_csp_long + sp_gain;
-
 	/* Check for overflow */
-
-	/* new_csp = LONG_MAX LONG_MIN;} DAVIDTODO produces warning*/
-	if ((new_csp_long < 0) && (old_csp_long > 0) && (sp_gain > 0)) {
-		new_csp_long = INT32_MAX;
-		sp_gain = 0;
-	} else if ((new_csp_long > 0) && (old_csp_long < 0) && (sp_gain < 0)) {
-		new_csp_long = INT32_MIN;
+	if (sp_gain > 0) {
+		if (old_32 < INT32_MAX - sp_gain) {
+			new_32 = old_32 + sp_gain;
+		} else {
+			new_32 = INT32_MAX;
+			sp_gain = 0;
+		}
+	} else if (old_32 > INT32_MIN - sp_gain) {
+		new_32 = old_32 + sp_gain;
+	} else {
+		new_32 = INT32_MIN;
 		sp_gain = 0;
 	}
 
 	/* Break it back down*/
-	p->csp = (s16b)(new_csp_long >> 16);   /* div 65536 */
-	p->csp_frac = (u16b)(new_csp_long & 0xFFFF);    /* mod 65536 */
+	if (new_32 < 0) {
+		/*
+		 * Don't use right bitwise shift on negative values:  whether
+		 * the left bits are zero or one depends on the system.
+		 */
+		int32_t remainder = new_32 % 65536;
+
+		p->csp = (int16_t) (new_32 / 65536);
+		if (remainder) {
+			assert(remainder < 0);
+			p->csp_frac = (uint16_t) (65536 + remainder);
+			assert(p->csp > INT16_MIN);
+			p->csp -= 1;
+		} else {
+			p->csp_frac = 0;
+		}
+	} else {
+		p->csp = (int16_t)(new_32 >> 16);   /* div 65536 */
+		p->csp_frac = (uint16_t)(new_32 & 0xFFFF);    /* mod 65536 */
+	}
 
 	/* Max/min SP */
 	if (p->csp >= p->msp) {
@@ -492,32 +639,32 @@ s32b player_adjust_mana_precise(struct player *p, s32b sp_gain)
 	}
 
 	/* Notice changes */
-	if (old_csp_short != p->csp) {
+	if (old_16 != p->csp) {
 		p->upkeep->redraw |= (PR_MANA);
 	}
 
 	if (sp_gain == 0) {
 		/* Recalculate */
-		new_csp_long = (s32b)((p->csp << 16) + p->csp_frac);
-		sp_gain = new_csp_long - old_csp_long;
+		new_32 = ((int32_t) p->csp) * 65536 + p->csp_frac;
+		sp_gain = new_32 - old_32;
 	}
 
 	return sp_gain;
 }
 
-void convert_mana_to_hp(struct player *p, s32b sp_long) {
-	s32b hp_gain, sp_ratio;
+void convert_mana_to_hp(struct player *p, int32_t sp_long) {
+	int32_t hp_gain, sp_ratio;
 
 	if (sp_long <= 0 || p->msp == 0 || p->mhp == p->chp) return;
 
 	/* Total HP from max */
-	hp_gain = (s32b)((p->mhp - p->chp) << 16);
-	hp_gain -= (s32b)p->chp_frac;
+	hp_gain = ((int32_t)(p->mhp - p->chp)) * 65536;
+	hp_gain -= (int32_t)p->chp_frac;
 
 	/* Spend X% of SP get X/2% of lost HP. E.g., at 50% HP get X/4% */
 	/* Gain stays low at msp<10 because MP gains are generous at msp<10 */
 	/* sp_ratio is max sp to spent sp, doubled to suit target rate. */
-	sp_ratio = (MAX(10, (s32b)p->msp) << 16) * 2 / sp_long;
+	sp_ratio = (((int32_t)MAX(10, (int32_t)p->msp)) * 131072) / sp_long;
 
 	/* Limit max healing to 25% of damage; ergo spending > 50% msp
 	 * is inefficient */
@@ -554,14 +701,14 @@ void player_update_light(struct player *p)
 			/* Decrease life-span */
 			obj->timeout--;
 
-			/* Hack -- notice interesting fuel steps */
+			/* Notice interesting fuel steps */
 			if ((obj->timeout < 100) || (!(obj->timeout % 100)))
 				/* Redraw stuff */
 				p->upkeep->redraw |= (PR_EQUIP);
 
-			/* Hack -- Special treatment when blind */
+			/* Special treatment when blind */
 			if (p->timed[TMD_BLIND]) {
-				/* Hack -- save some light for later */
+				/* Save some light for later */
 				if (obj->timeout == 0) obj->timeout++;
 			} else if (obj->timeout == 0) {
 				/* The light is now out */
@@ -571,11 +718,12 @@ void player_update_light(struct player *p)
 				/* If it's a torch, now is the time to delete it */
 				if (of_has(obj->flags, OF_BURNS_OUT)) {
 					bool dummy;
-					struct object *burnt = gear_object_for_use(obj, 1, false,
-															   &dummy);
+					struct object *burnt =
+						gear_object_for_use(p, obj, 1,
+						false, &dummy);
 					if (burnt->known)
-						object_delete(&burnt->known);
-					object_delete(&burnt);
+						object_delete(p->cave, NULL, &burnt->known);
+					object_delete(cave, p->cave, &burnt);
 				}
 			} else if ((obj->timeout < 50) && (!(obj->timeout % 20))) {
 				/* The light is getting dim */
@@ -595,23 +743,42 @@ void player_update_light(struct player *p)
  */
 struct object *player_best_digger(struct player *p, bool forbid_stack)
 {
+	int weapon_slot = slot_by_name(p, "weapon");
+	struct object *current_weapon = slot_object(p, weapon_slot);
 	struct object *obj, *best = NULL;
-	int best_score = 0;
+	/* Prefer any melee weapon over unarmed digging, i.e. best == NULL. */
+	int best_score = -1;
+	struct player_state local_state;
 
 	for (obj = p->gear; obj; obj = obj->next) {
-		int score = 0;
+		int score, old_number;
 		if (!tval_is_melee_weapon(obj)) continue;
 		if (obj->number < 1 || (forbid_stack && obj->number > 1)) continue;
-		if (tval_is_digger(obj)) {
-			if (of_has(obj->flags, OF_DIG_1))
-				score = 1;
-			else if (of_has(obj->flags, OF_DIG_2))
-				score = 2;
-			else if (of_has(obj->flags, OF_DIG_3))
-				score = 3;
+		/* Don't use it if it has a sticky curse. */
+		if (!obj_can_takeoff(obj)) continue;
+
+		/* Swap temporarily for the calc_bonuses() computation. */
+		old_number = obj->number;
+		if (obj != current_weapon) {
+			obj->number = 1;
+			p->body.slots[weapon_slot].obj = obj;
 		}
-		score += obj->modifiers[OBJ_MOD_TUNNEL]
-			* p->obj_k->modifiers[OBJ_MOD_TUNNEL];
+
+		/*
+		 * Avoid side effects from using update set to false
+		 * with calc_bonuses().
+		 */
+		local_state.stat_ind[STAT_STR] = 0;
+		local_state.stat_ind[STAT_DEX] = 0;
+		calc_bonuses(p, &local_state, true, false);
+		score = local_state.skills[SKILL_DIGGING];
+
+		/* Swap back. */
+		if (obj != current_weapon) {
+			obj->number = old_number;
+			p->body.slots[weapon_slot].obj = current_weapon;
+		}
+
 		if (score > best_score) {
 			best = obj;
 			best_score = score;
@@ -633,11 +800,12 @@ bool player_attack_random_monster(struct player *p)
 
 	/* Look for a monster, attack */
 	for (i = 0; i < 8; i++, dir++) {
-		struct loc grid = loc_sum(player->grid, ddgrid_ddd[dir % 8]);
-		if (square_monster(cave, grid)) {
+		struct loc grid = loc_sum(p->grid, ddgrid_ddd[dir % 8]);
+		const struct monster *mon = square_monster(cave, grid);
+		if (mon && !monster_is_camouflaged(mon)) {
 			p->upkeep->energy_use = z_info->move_energy;
 			msg("You angrily lash out at a nearby foe!");
-			move_player(ddd[dir % 8], false);
+			py_attack(p, grid);
 			return true;
 		}
 	}
@@ -669,16 +837,16 @@ void player_over_exert(struct player *p, int flag, int chance, int amount)
 			msg("You faint from the effort!");
 
 			/* Bypass free action */
-			(void)player_inc_timed(p, TMD_PARALYZED, randint1(amount),
-								   true, false);
+			(void)player_inc_timed(p, TMD_PARALYZED,
+				randint1(amount), true, true, false);
 		}
 	}
 
 	/* Scrambled stats */
 	if (flag & PY_EXERT_SCRAMBLE) {
 		if (randint0(100) < chance) {
-			(void)player_inc_timed(p, TMD_SCRAMBLE, randint1(amount),
-								   true, true);
+			(void)player_inc_timed(p, TMD_SCRAMBLE,
+				randint1(amount), true, true, true);
 		}
 	}
 
@@ -687,15 +855,15 @@ void player_over_exert(struct player *p, int flag, int chance, int amount)
 		if (randint0(100) < chance) {
 			msg("Wounds appear on your body!");
 			(void)player_inc_timed(p, TMD_CUT, randint1(amount),
-								   true, false);
+				true, true, false);
 		}
 	}
 
 	/* Confusion */
 	if (flag & PY_EXERT_CONF) {
 		if (randint0(100) < chance) {
-			(void)player_inc_timed(p, TMD_CONFUSED, randint1(amount),
-								   true, true);
+			(void)player_inc_timed(p, TMD_CONFUSED,
+				randint1(amount), true, true, true);
 		}
 	}
 
@@ -703,7 +871,7 @@ void player_over_exert(struct player *p, int flag, int chance, int amount)
 	if (flag & PY_EXERT_HALLU) {
 		if (randint0(100) < chance) {
 			(void)player_inc_timed(p, TMD_IMAGE, randint1(amount),
-								   true, true);
+				true, true, true);
 		}
 	}
 
@@ -712,24 +880,37 @@ void player_over_exert(struct player *p, int flag, int chance, int amount)
 		if (randint0(100) < chance) {
 			msg("You feel suddenly lethargic.");
 			(void)player_inc_timed(p, TMD_SLOW, randint1(amount),
-								   true, false);
+				true, true, false);
 		}
 	}
 
 	/* HP */
 	if (flag & PY_EXERT_HP) {
 		if (randint0(100) < chance) {
-			msg("You cry out in sudden pain!");
-			take_hit(p, randint1(amount), "over-exertion");
+			int dam = player_apply_damage_reduction(p,
+				randint1(amount));
+			char dam_text[32] = "";
+
+			if (dam > 0 && OPT(p, show_damage)) {
+				strnfmt(dam_text, sizeof(dam_text),
+					" (%d)", dam);
+			}
+			msg("You cry out in sudden pain!%s", dam_text);
+			take_hit(p, dam, "over-exertion");
 		}
 	}
 }
 
 
 /**
- * See how much damage the player will take from damaging terrain
+ * See how much damage the player will take from terrain.
+ *
+ * \param p is the player to check
+ * \param grid is the location of the terrain
+ * \param actual will, if true, cause the player to learn the appropriate
+ * runes if equipment or effects mitigate the damage.
  */
-int player_check_terrain_damage(struct player *p, struct loc grid)
+int player_check_terrain_damage(struct player *p, struct loc grid, bool actual)
 {
 	int dam_taken = 0;
 
@@ -738,11 +919,15 @@ int player_check_terrain_damage(struct player *p, struct loc grid)
 		int res = p->state.el_info[ELEM_FIRE].res_level;
 
 		/* Fire damage */
-		dam_taken = adjust_dam(p, ELEM_FIRE, base_dam, RANDOMISE, res, false);
+		dam_taken = adjust_dam(p, ELEM_FIRE, base_dam, RANDOMISE, res,
+			actual);
 
 		/* Feather fall makes one lightfooted. */
 		if (player_of_has(p, OF_FEATHER)) {
 			dam_taken /= 2;
+			if (actual) {
+				equip_learn_flag(p, OF_FEATHER);
+			}
 		}
 	}
 
@@ -754,18 +939,30 @@ int player_check_terrain_damage(struct player *p, struct loc grid)
  */
 void player_take_terrain_damage(struct player *p, struct loc grid)
 {
-	int dam_taken = player_check_terrain_damage(p, grid);
+	int dam_taken = player_check_terrain_damage(p, grid, true);
+	int dam_reduced;
 
 	if (!dam_taken) {
 		return;
 	}
 
-	/* Damage the player and inventory */
-	take_hit(player, dam_taken, square_feat(cave, grid)->die_msg);
+	/*
+	 * Damage the player and inventory; inventory damage is based on
+	 * the raw incoming damage and not the value accounting for the
+	 * player's damage reduction.
+	 */
+	dam_reduced = player_apply_damage_reduction(p, dam_taken);
 	if (square_isfiery(cave, grid)) {
-		msg(square_feat(cave, grid)->hurt_msg);
-		inven_damage(player, PROJ_FIRE, dam_taken);
+		char dam_text[32] = "";
+
+		if (dam_reduced > 0 && OPT(p, show_damage)) {
+			strnfmt(dam_text, sizeof(dam_text), " (%d)",
+				dam_reduced);
+		}
+		msg("%s%s", square_feat(cave, grid)->hurt_msg, dam_text);
+		inven_damage(p, PROJ_FIRE, dam_taken);
 	}
+	take_hit(p, dam_reduced, square_feat(cave, grid)->die_msg);
 }
 
 /**
@@ -813,6 +1010,38 @@ struct player_shape *player_shape_by_idx(int index)
 	return NULL;
 }
 
+/**
+ * Give shapechanged players a choice of returning to normal shape and
+ * performing a command, just returning to normal shape without acting, or
+ * canceling.
+ *
+ * \param p the player
+ * \param cmd the command being performed
+ * \return true if the player wants to proceed with their command
+ */
+bool player_get_resume_normal_shape(struct player *p, struct command *cmd)
+{
+	if (player_is_shapechanged(p)) {
+		msg("You cannot do this while in %s form.", p->shape->name);
+		char prompt[100];
+		strnfmt(prompt, sizeof(prompt),
+		        "Change back and %s (y/n) or (r)eturn to normal? ",
+		        cmd_verb(cmd->code));
+		char answer = get_char(prompt, "yrn", 3, 'n');
+
+		// Change back to normal shape
+		if (answer == 'y' || answer == 'r') {
+			player_resume_normal_shape(p);
+		}
+
+		// Players may only act if they return to normal shape
+		return answer == 'y';
+	}
+
+	// Normal shape players can proceed as usual
+	return true;
+}
+
 /**
  * Revert to normal shape
  */
@@ -822,18 +1051,18 @@ void player_resume_normal_shape(struct player *p)
 	msg("You resume your usual shape.");
 
 	/* Kill vampire attack */
-	(void) player_clear_timed(p, TMD_ATT_VAMP, true);
+	(void) player_clear_timed(p, TMD_ATT_VAMP, true, false);
 
 	/* Update */
-	player->upkeep->update |= (PU_BONUS);
-	player->upkeep->redraw |= (PR_TITLE | PR_MISC);
-	handle_stuff(player);
+	p->upkeep->update |= (PU_BONUS);
+	p->upkeep->redraw |= (PR_TITLE | PR_MISC);
+	handle_stuff(p);
 }
 
 /**
  * Check if the player is shapechanged
  */
-bool player_is_shapechanged(struct player *p)
+bool player_is_shapechanged(const struct player *p)
 {
 	return streq(p->shape->name, "normal") ? false : true;
 }
@@ -841,7 +1070,7 @@ bool player_is_shapechanged(struct player *p)
 /**
  * Check if the player is immune from traps
  */
-bool player_is_trapsafe(struct player *p)
+bool player_is_trapsafe(const struct player *p)
 {
 	if (p->timed[TMD_TRAPSAFE]) return true;
 	if (player_of_has(p, OF_TRAP_IMMUNE)) return true;
@@ -855,7 +1084,7 @@ bool player_is_trapsafe(struct player *p)
  * \param show_msg should be set to true if a failure message should be
  * displayed.
  */
-bool player_can_cast(struct player *p, bool show_msg)
+bool player_can_cast(const struct player *p, bool show_msg)
 {
 	if (!p->class->magic.total_spells) {
 		if (show_msg) {
@@ -864,7 +1093,7 @@ bool player_can_cast(struct player *p, bool show_msg)
 		return false;
 	}
 
-	if (p->timed[TMD_BLIND] || no_light()) {
+	if (p->timed[TMD_BLIND] || no_light(p)) {
 		if (show_msg) {
 			msg("You cannot see!");
 		}
@@ -888,7 +1117,7 @@ bool player_can_cast(struct player *p, bool show_msg)
  * \param show_msg should be set to true if a failure message should be
  * displayed.
  */
-bool player_can_study(struct player *p, bool show_msg)
+bool player_can_study(const struct player *p, bool show_msg)
 {
 	if (!player_can_cast(p, show_msg))
 		return false;
@@ -934,7 +1163,7 @@ bool player_can_study(struct player *p, bool show_msg)
  * \param show_msg should be set to true if a failure message should be
  * displayed.
  */
-bool player_can_read(struct player *p, bool show_msg)
+bool player_can_read(const struct player *p, bool show_msg)
 {
 	if (p->timed[TMD_BLIND]) {
 		if (show_msg)
@@ -943,7 +1172,7 @@ bool player_can_read(struct player *p, bool show_msg)
 		return false;
 	}
 
-	if (no_light()) {
+	if (no_light(p)) {
 		if (show_msg)
 			msg("You have no light to read by.");
 
@@ -1011,7 +1240,8 @@ bool player_can_refuel(struct player *p, bool show_msg)
 }
 
 /**
- * Prerequiste function for command. See struct cmd_info in cmd-process.c.
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
  */
 bool player_can_cast_prereq(void)
 {
@@ -1019,7 +1249,8 @@ bool player_can_cast_prereq(void)
 }
 
 /**
- * Prerequiste function for command. See struct cmd_info in cmd-process.c.
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
  */
 bool player_can_study_prereq(void)
 {
@@ -1027,15 +1258,22 @@ bool player_can_study_prereq(void)
 }
 
 /**
- * Prerequiste function for command. See struct cmd_info in cmd-process.c.
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
  */
 bool player_can_read_prereq(void)
 {
-	return player_can_read(player, true);
+	/*
+	 * Accommodate hacks elsewhere:  'r' is overloaded to mean
+	 * release a commanded monster when TMD_COMMAND is active.
+	 */
+	return (player->timed[TMD_COMMAND]) ?
+		true : player_can_read(player, true);
 }
 
 /**
- * Prerequiste function for command. See struct cmd_info in cmd-process.c.
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
  */
 bool player_can_fire_prereq(void)
 {
@@ -1043,13 +1281,32 @@ bool player_can_fire_prereq(void)
 }
 
 /**
- * Prerequiste function for command. See struct cmd_info in cmd-process.c.
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
  */
 bool player_can_refuel_prereq(void)
 {
 	return player_can_refuel(player, true);
 }
 
+/**
+ * Prerequisite function for command. See struct cmd_info in ui-input.h and
+ * it's use in ui-game.c.
+ */
+bool player_can_debug_prereq(void)
+{
+	if (player->noscore & NOSCORE_DEBUG) {
+		return true;
+	}
+	if (confirm_debug()) {
+		/* Mark savefile */
+		player->noscore |= NOSCORE_DEBUG;
+		return true;
+	}
+	return false;
+}
+
+
 /**
  * Return true if the player has access to a book that has unlearned spells.
  *
@@ -1069,15 +1326,15 @@ bool player_book_has_unlearned_spells(struct player *p)
 	}
 
 	/* Check through all available books */
-	item_num = scan_items(item_list, item_max, USE_INVEN | USE_FLOOR,
-						  obj_can_study);
+	item_num = scan_items(item_list, item_max, p, USE_INVEN | USE_FLOOR,
+		obj_can_study);
 	for (i = 0; i < item_num; i++) {
 		const struct class_book *book = player_object_to_book(p, item_list[i]);
 		if (!book) continue;
 
 		/* Extract spells */
 		for (j = 0; j < book->num_spells; j++)
-			if (spell_okay_to_study(book->spells[j].sidx)) {
+			if (spell_okay_to_study(p, book->spells[j].sidx)) {
 				/* There is a spell the player can study */
 				mem_free(item_list);
 				return true;
@@ -1122,7 +1379,7 @@ bool player_confuse_dir(struct player *p, int *dp, bool too)
 /**
  * Return true if the provided count is one of the conditional REST_ flags.
  */
-bool player_resting_is_special(s16b count)
+bool player_resting_is_special(int16_t count)
 {
 	switch (count) {
 		case REST_COMPLETE:
@@ -1137,7 +1394,7 @@ bool player_resting_is_special(s16b count)
 /**
  * Return true if the player is resting.
  */
-bool player_is_resting(struct player *p)
+bool player_is_resting(const struct player *p)
 {
 	return (p->upkeep->resting > 0 ||
 			player_resting_is_special(p->upkeep->resting));
@@ -1146,7 +1403,7 @@ bool player_is_resting(struct player *p)
 /**
  * Return the remaining number of resting turns.
  */
-s16b player_resting_count(struct player *p)
+int16_t player_resting_count(const struct player *p)
 {
 	return p->upkeep->resting;
 }
@@ -1162,9 +1419,10 @@ static bool player_rest_disturb = false;
 /**
  * Set the number of resting turns.
  *
+ * \param p is the player trying to rest.
  * \param count is the number of turns to rest or one of the REST_ constants.
  */
-void player_resting_set_count(struct player *p, s16b count)
+void player_resting_set_count(struct player *p, int16_t count)
 {
 	/* Cancel if player is disturbed */
 	if (player_rest_disturb) {
@@ -1200,7 +1458,7 @@ void player_resting_cancel(struct player *p, bool disturb)
  * Return true if the player should get a regeneration bonus for the current
  * rest.
  */
-bool player_resting_can_regenerate(struct player *p)
+bool player_resting_can_regenerate(const struct player *p)
 {
 	return player_turns_rested >= REST_REQUIRED_FOR_REGEN ||
 		player_resting_is_special(p->upkeep->resting);
@@ -1277,9 +1535,10 @@ int player_get_resting_repeat_count(struct player *p)
 /**
  * Set the number of resting turns to repeat.
  *
+ * \param p is the player trying to rest.
  * \param count is the number of turns requested for rest most recently.
  */
-void player_set_resting_repeat_count(struct player *p, s16b count)
+void player_set_resting_repeat_count(struct player *p, int16_t count)
 {
 	player_resting_repeat_count = count;
 }
@@ -1287,7 +1546,7 @@ void player_set_resting_repeat_count(struct player *p, s16b count)
 /**
  * Check if the player state has the given OF_ flag.
  */
-bool player_of_has(struct player *p, int flag)
+bool player_of_has(const struct player *p, int flag)
 {
 	assert(p);
 	return of_has(p->state.flags, flag);
@@ -1296,7 +1555,7 @@ bool player_of_has(struct player *p, int flag)
 /**
  * Check if the player resists (or better) an element
  */
-bool player_resists(struct player *p, int element)
+bool player_resists(const struct player *p, int element)
 {
 	return (p->state.el_info[element].res_level > 0);
 }
@@ -1304,7 +1563,7 @@ bool player_resists(struct player *p, int element)
 /**
  * Check if the player resists (or better) an element
  */
-bool player_is_immune(struct player *p, int element)
+bool player_is_immune(const struct player *p, int element)
 {
 	return (p->state.el_info[element].res_level == 3);
 }
@@ -1327,13 +1586,57 @@ void player_place(struct chunk *c, struct player *p, struct loc grid)
 	p->upkeep->create_up_stair = false;
 }
 
+/*
+ * Take care of bookkeeping after moving the player with monster_swap().
+ *
+ * \param p is the player that was moved.
+ * \param eval_trap will, if true, cause evaluation (possibly affecting the
+ * player) of the traps in the grid.
+ * \param is_involuntary will, if true, do appropriate actions (flush the
+ * command queue) for a move not expected by the player.
+ */
+void player_handle_post_move(struct player *p, bool eval_trap,
+		bool is_involuntary)
+{
+	/* Handle store doors, or notice objects */
+	if (square_isshop(cave, p->grid)) {
+		if (player_is_shapechanged(p)) {
+			if (square(cave, p->grid)->feat != FEAT_HOME) {
+				msg("There is a scream and the door slams shut!");
+			}
+			return;
+		}
+		disturb(p);
+		if (is_involuntary) {
+			cmdq_flush();
+		}
+		event_signal(EVENT_ENTER_STORE);
+		event_remove_handler_type(EVENT_ENTER_STORE);
+		event_signal(EVENT_USE_STORE);
+		event_remove_handler_type(EVENT_USE_STORE);
+		event_signal(EVENT_LEAVE_STORE);
+		event_remove_handler_type(EVENT_LEAVE_STORE);
+	} else {
+		if (is_involuntary) {
+			cmdq_flush();
+		}
+		square_know_pile(cave, p->grid, NULL);
+	}
+
+	/* Discover invisible traps, set off visible ones */
+	if (eval_trap && square_isplayertrap(cave, p->grid)
+			&& !square_isdisabledtrap(cave, p->grid)) {
+		hit_trap(p->grid, 0);
+	}
+
+	/* Update view and search */
+	update_view(cave, p);
+	search(p);
+}
+
 /*
  * Something has happened to disturb the player.
  *
- * The first arg indicates a major disturbance, which affects search.
- *
- * The second arg is currently unused, but could induce output flush.
- *
  * All disturbance cancels repeated commands, resting, and running.
  *
  * XXX-AS: Make callers either pass in a command
@@ -1353,6 +1656,8 @@ void disturb(struct player *p)
 	/* Cancel running */
 	if (p->upkeep->running) {
 		p->upkeep->running = 0;
+		mem_free(p->upkeep->steps);
+		p->upkeep->steps = NULL;
 
 		/* Cancel queued commands */
 		cmdq_flush();
@@ -1377,7 +1682,7 @@ void search(struct player *p)
 	struct loc grid;
 
 	/* Various conditions mean no searching */
-	if (p->timed[TMD_BLIND] || no_light() ||
+	if (p->timed[TMD_BLIND] || no_light(p) ||
 		p->timed[TMD_CONFUSED] || p->timed[TMD_IMAGE])
 		return;
 
@@ -1395,8 +1700,10 @@ void search(struct player *p)
 
 			/* Traps on chests */
 			for (obj = square_object(cave, grid); obj; obj = obj->next) {
-				if (!obj->known || !is_trapped_chest(obj))
+				if (!obj->known || ignore_item_ok(p, obj)
+						|| !is_trapped_chest(obj)) {
 					continue;
+				}
 
 				if (obj->known->pval != obj->pval) {
 					msg("You have discovered a trap on the chest!");
@@ -1407,3 +1714,20 @@ void search(struct player *p)
 		}
 	}
 }
+
+/**
+ * Test if there are any monsters the player knows about in the field of view.
+ */
+bool player_has_monster_in_view(const struct player *p)
+{
+	int n = cave_monster_max(cave), i;
+
+	for (i = 1; i < n; ++i) {
+		const struct monster *mon = cave_monster(cave, i);
+
+		if (monster_is_obvious(mon) && monster_is_in_view(mon)) {
+			return true;
+		}
+	}
+	return false;
+}
