@@ -28,6 +28,7 @@
 #include "player-path.h"
 #include "savefile.h"
 #include "target.h"
+#include "ui-birth.h"
 #include "ui-command.h"
 #include "ui-context.h"
 #include "ui-curse.h"
@@ -40,6 +41,7 @@
 #include "ui-menu.h"
 #include "ui-object.h"
 #include "ui-output.h"
+#include "ui-player-properties.h"
 #include "ui-player.h"
 #include "ui-prefs.h"
 #include "ui-signals.h"
@@ -115,6 +117,17 @@ struct keypress *inkey_next = NULL;
  */
 static bool keymap_auto_more;
 
+#ifdef ALLOW_BORG
+
+/*
+ * Mega-Hack -- special "inkey_hack" hook.  XXX XXX XXX
+ *
+ * This special function hook allows the "Borg" (see elsewhere) to take
+ * control of the "inkey()" function, and substitute in fake keypresses.
+ */
+struct keypress(*inkey_hack)(int flush_first) = NULL;
+
+#endif /* ALLOW_BORG */
 
 /**
  * Get a keypress from the user.
@@ -168,7 +181,7 @@ ui_event inkey_ex(void)
 		inkey_xtra = false;
 	}
 
-	/* Hack -- Use the "inkey_next" pointer */
+	/* Use the "inkey_next" pointer */
 	while (inkey_next && inkey_next->code) {
 		/* Get next character, and advance */
 		ke.key = *inkey_next++;
@@ -208,6 +221,24 @@ ui_event inkey_ex(void)
 	/* Forget pointer */
 	inkey_next = NULL;
 
+#ifdef ALLOW_BORG
+	/* Mega-Hack -- Use the special hook */
+	if (inkey_hack)
+	{
+		ke.key = (*inkey_hack)(inkey_xtra);
+		if (ke.key.type != EVT_NONE)
+		{
+			/* Cancel the various "global parameters" */
+			inkey_flag = false;
+			inkey_scan = 0;
+			ke.type = EVT_KBRD;
+
+			/* Accept result */
+			return (ke);
+		}
+	}
+#endif /* ALLOW_BORG */
+
 	/* Get the cursor state */
 	(void)Term_get_cursor(&cursor_state);
 
@@ -217,27 +248,27 @@ ui_event inkey_ex(void)
 		(void)Term_set_cursor(true);
 
 
-	/* Hack -- Activate main screen */
+	/* Activate main screen */
 	Term_activate(term_screen);
 
 
 	/* Get a key */
 	while (ke.type == EVT_NONE) {
-		/* Hack -- Handle "inkey_scan == SCAN_INSTANT */
+		/* Handle "inkey_scan == SCAN_INSTANT */
 		if (inkey_scan == SCAN_INSTANT &&
 			(0 != Term_inkey(&kk, false, false)))
 			break;
 
 
-		/* Hack -- Flush output once when no key ready */
+		/* Flush output once when no key ready */
 		if (!done && (0 != Term_inkey(&kk, false, false))) {
-			/* Hack -- activate proper term */
+			/* Activate proper term */
 			Term_activate(old);
 
 			/* Flush output */
 			Term_fresh();
 
-			/* Hack -- activate main screen */
+			/* Activate main screen */
 			Term_activate(term_screen);
 
 			/* Mega-Hack -- reset saved flag */
@@ -263,7 +294,7 @@ ui_event inkey_ex(void)
 			ke.key.code = ESCAPE;
 	}
 
-	/* Hack -- restore the term */
+	/* Restore the term */
 	Term_activate(old);
 
 	/* Restore the cursor */
@@ -349,7 +380,7 @@ ui_event inkey_m(void)
 
 
 /**
- * Hack -- flush
+ * Flush
  */
 static void msg_flush(int x)
 {
@@ -365,6 +396,58 @@ static void msg_flush(int x)
 	Term_erase(0, 0, 255);
 }
 
+/**
+ * Like msg_flush() but split what has already been pushed to the Term's
+ * buffer to make room for the "-more-" prompt.
+ *
+ * \param w is the number of columns in the terminal
+ * \param x points to the integer storing the column where the next
+ * message will start.
+ */
+static void msg_flush_split_existing(int w, int *x)
+{
+	/* Default place to split what's there */
+	int split = MIN(*x, w - 8);
+	int i = split;
+	wchar_t *svc = NULL;
+	int *sva = NULL;
+
+	/* Find the rightmost split point. */
+	while (i > w / 2) {
+		int a;
+		wchar_t c;
+
+		--i;
+		Term_what(i, 0, &a, &c);
+		if (c == L' ') {
+			split = i;
+			break;
+		}
+	}
+
+	/* Remember what's on and after the split point. */
+	*x -= split;
+	if (*x > 0) {
+		svc = mem_alloc(*x * sizeof(*svc));
+		sva = mem_alloc(*x * sizeof(*sva));
+		for (i = 0; i < *x; ++i) {
+			Term_what(i + split, 0, &sva[i], &svc[i]);
+		}
+	}
+
+	Term_erase(split, 0, w);
+	msg_flush(split + 1);
+
+	/* Put back what was remembered. */
+	if (*x > 0) {
+		for (i = 0; i < *x; ++i) {
+			Term_putch(i, 0, sva[i], svc[i]);
+		}
+		mem_free(sva);
+		mem_free(svc);
+	}
+}
+
 static int message_column = 0;
 
 
@@ -395,7 +478,7 @@ bool msg_flag;
  * result in the loss of information if the screen is cleared, or if anything
  * is displayed on the top line.
  *
- * Hack -- Note that "msg("%s", NULL)" will clear the top line even if no
+ * Note that "msg("%s", NULL)" will clear the top line even if no
  * messages are pending.
  */
 void display_message(game_event_type unused, game_event_data *data, void *user)
@@ -425,22 +508,24 @@ void display_message(game_event_type unused, game_event_data *data, void *user)
 	/* Obtain the size */
 	(void)Term_get_size(&w, &h);
 
-	/* Hack -- Reset */
+	/* Reset */
 	if (!msg_flag) message_column = 0;
 
 	/* Message Length */
 	n = (msg ? strlen(msg) : 0);
 
-	/* Hack -- flush when requested or needed */
+	/* Flush when requested or needed */
 	if (message_column && (!msg || ((message_column + n) > (w - 8)))) {
 		/* Flush */
-		msg_flush(message_column);
+		if (message_column <= w - 8) {
+			msg_flush(message_column);
+			message_column = 0;
+		} else {
+			msg_flush_split_existing(w, &message_column);
+		}
 
 		/* Forget it */
 		msg_flag = false;
-
-		/* Reset */
-		message_column = 0;
 	}
 
 	/* No message */
@@ -459,17 +544,20 @@ void display_message(game_event_type unused, game_event_data *data, void *user)
 	color = message_type_color(type);
 
 	/* Split message */
-	while (n > w - 1) {
+	while (message_column + n > w - 1) {
+		/* Default split */
+		int split = MAX(w - 8 - message_column, 0);
+		int check = split;
 		char oops;
 
-		int check, split;
-
-		/* Default split */
-		split = w - 8;
-
 		/* Find the rightmost split point */
-		for (check = (w / 2); check < w - 8; check++)
-			if (t[check] == ' ') split = check;
+		while (check > MAX(w / 2 - message_column, 0)) {
+			--check;
+			if (t[check] == ' ') {
+				split = check;
+				break;
+			}
+		}
 
 		/* Save the split character */
 		oops = t[split];
@@ -478,10 +566,10 @@ void display_message(game_event_type unused, game_event_data *data, void *user)
 		t[split] = '\0';
 
 		/* Display part of the message */
-		Term_putstr(0, 0, split, color, t);
+		Term_putstr(message_column, 0, split, color, t);
 
 		/* Flush it */
-		msg_flush(split + 1);
+		msg_flush(message_column + split + 1);
 
 		/* Restore the split character */
 		t[split] = oops;
@@ -490,7 +578,7 @@ void display_message(game_event_type unused, game_event_data *data, void *user)
 		t[--split] = ' ';
 
 		/* Prepare to recurse on the rest of "buf" */
-		t += split; n -= split;
+		t += split; n -= split; message_column = 0;
 	}
 
 	/* Display the tail of the message */
@@ -520,14 +608,23 @@ void bell_message(game_event_type unused, game_event_data *data, void *user)
  */
 void message_flush(game_event_type unused, game_event_data *data, void *user)
 {
-	/* Hack -- Reset */
+	/* Reset */
 	if (!msg_flag) message_column = 0;
 
 	/* Flush when needed */
 	if (message_column) {
 		/* Print pending messages */
-		if (Term)
-			msg_flush(message_column);
+		if (Term) {
+			int w, h;
+
+			(void)Term_get_size(&w, &h);
+			while (message_column > w - 8) {
+				msg_flush_split_existing(w, &message_column);
+			}
+			if (message_column) {
+				msg_flush(message_column);
+			}
+		}
 
 		/* Forget it */
 		msg_flag = false;
@@ -1177,7 +1274,7 @@ static bool textui_get_check(const char *prompt)
 	char buf[80];
 
 	/*
-	 * Hack -- Build a "useful" prompt; do this first so prompts built by
+	 * Build a "useful" prompt; do this first so prompts built by
 	 * format() won't run afoul of event_signal()'s side effects.
 	 */
 	strnfmt(buf, 78, "%.70s[y/n] ", prompt);
@@ -1223,7 +1320,7 @@ char get_char(const char *prompt, const char *options, size_t len, char fallback
 	/* Paranoia */
 	event_signal(EVENT_MESSAGE_FLUSH);
 
-	/* Hack -- Build a "useful" prompt */
+	/* Build a "useful" prompt */
 	strnfmt(buf, 78, "%.70s[%s] ", prompt, options);
 
 	/* Prompt for it */
@@ -1390,19 +1487,17 @@ static int dir_transitions[10][10] =
 };
 
 /**
- * Request a "movement" direction (1,2,3,4,6,7,8,9) from the user.
+ * Request a "movement" direction (1,2,3,4,5(optional),6,7,8,9) from the user.
  *
  * Return true if a direction was chosen, otherwise return false.
  *
  * This function should be used for all "repeatable" commands, such as
  * run, walk, open, close, bash, disarm, spike, tunnel, etc, as well
- * as all commands which must reference a grid adjacent to the player,
- * and which may not reference the grid under the player.
+ * as all commands which must reference a grid adjacent to the player.
+ * If the command does not allow the grid under the player, pass false
+ * for allow_5.  Otherwise, use true for allow_5.
  *
- * Directions "5" and "0" are illegal and will not be accepted.
- *
- * This function tracks and uses the "global direction", and uses
- * that as the "desired direction", if it is set.
+ * The direction, "0", is illegal and will not be accepted.
  */
 static bool textui_get_rep_dir(int *dp, bool allow_5)
 {
@@ -1427,7 +1522,8 @@ static bool textui_get_rep_dir(int *dp, bool allow_5)
 		inkey_scan = SCAN_OFF;
 
 		if (ke.type == EVT_NONE ||
-			(ke.type == EVT_KBRD && target_dir(ke.key) == 0)) {
+				(ke.type == EVT_KBRD
+				&& !target_dir_allow(ke.key, allow_5, true))) {
 			prt("Direction or <click> (Escape to cancel)? ", 0, 0);
 			ke = inkey_ex();
 		}
@@ -1462,10 +1558,17 @@ static bool textui_get_rep_dir(int *dp, bool allow_5)
 
 				/* XXX Ideally show and move the cursor here to indicate
 				 the currently "Pending" direction. XXX */
-				this_dir = target_dir_allow(ke.key, allow_5);
+				this_dir = target_dir_allow(ke.key, allow_5,
+					true);
 
-				if (this_dir)
+				if (this_dir == ESCAPE) {
+					/* Clear the prompt */
+					prt("", 0, 0);
+
+					return (false);
+				} else if (this_dir) {
 					dir = dir_transitions[dir][this_dir];
+				}
 
 				if (player->opts.lazymove_delay == 0 || ++keypresses_handled > 1)
 					break;
@@ -1514,7 +1617,6 @@ static bool textui_get_aim_dir(int *dp)
 {
 	/* Global direction */
 	int dir = 0;
-
 	ui_event ke;
 
 	const char *p;
@@ -1522,11 +1624,17 @@ static bool textui_get_aim_dir(int *dp)
 	/* Initialize */
 	(*dp) = 0;
 
-	/* Hack -- auto-target if requested */
+	/* Auto-target if requested */
 	if (OPT(player, use_old_target) && target_okay() && !dir) dir = 5;
 
 	/* Ask until satisfied */
 	while (!dir) {
+		/*
+		 * Whether to generate an audible warning about a targeting
+		 * failure.
+		 */
+		bool need_beep = false;
+
 		/* Choose a prompt */
 		if (!target_okay())
 			p = "Direction ('*' or <click> to target, \"'\" for closest, Escape to cancel)? ";
@@ -1538,8 +1646,9 @@ static bool textui_get_aim_dir(int *dp)
 
 		if (ke.type == EVT_MOUSE) {
 			if (ke.mouse.button == 1) {
-				if (target_set_interactive(TARGET_KILL, KEY_GRID_X(ke),
-										   KEY_GRID_Y(ke)))
+				if (target_set_interactive(TARGET_KILL,
+						KEY_GRID_X(ke), KEY_GRID_Y(ke),
+						false))
 					dir = 5;
 			} else if (ke.mouse.button == 2) {
 				break;
@@ -1547,16 +1656,23 @@ static bool textui_get_aim_dir(int *dp)
 		} else if (ke.type == EVT_KBRD) {
 			if (ke.key.code == '*') {
 				/* Set new target, use target if legal */
-				if (target_set_interactive(TARGET_KILL, -1, -1))
+				if (target_set_interactive(TARGET_KILL, -1, -1,
+						false))
 					dir = 5;
 			} else if (ke.key.code == '\'') {
 				/* Set to closest target */
-				if (target_set_closest(TARGET_KILL, NULL))
+				if (target_set_closest(TARGET_KILL, NULL)) {
 					dir = 5;
+				} else {
+					need_beep = true;
+				}
 			} else if (ke.key.code == 't' || ke.key.code == '5' ||
 					   ke.key.code == '0' || ke.key.code == '.') {
-				if (target_okay())
+				if (target_okay()) {
 					dir = 5;
+				} else {
+					need_beep = true;
+				}
 			} else {
 				/* Possible direction */
 				int keypresses_handled = 0;
@@ -1566,12 +1682,18 @@ static bool textui_get_aim_dir(int *dp)
 
 					/* XXX Ideally show and move the cursor here to indicate
 					 * the currently "Pending" direction. XXX */
-					this_dir = target_dir(ke.key);
+					this_dir = target_dir_allow(ke.key,
+						false, true);
 
-					if (this_dir)
+					if (this_dir == ESCAPE) {
+						return false;
+					}
+					if (this_dir) {
 						dir = dir_transitions[dir][this_dir];
-					else
+					} else {
+						need_beep = true;
 						break;
+					}
 
 					if (player->opts.lazymove_delay == 0 || ++keypresses_handled > 1)
 						break;
@@ -1585,7 +1707,7 @@ static bool textui_get_aim_dir(int *dp)
 		}
 
 		/* Error */
-		if (!dir) bell();
+		if (need_beep) bell();
 	}
 
 	/* No direction */
@@ -1617,6 +1739,7 @@ void textui_input_init(void)
 	get_panel_hook = textui_get_panel;
 	panel_contains_hook = textui_panel_contains;
 	map_is_visible_hook = textui_map_is_visible;
+	view_abilities_hook = textui_view_ability_menu;
 }
 
 
@@ -1667,7 +1790,7 @@ static int textui_get_count(void)
 
 
 /**
- * Hack -- special buffer to hold the action of the current keymap
+ * Special buffer to hold the action of the current keymap
  */
 static struct keypress request_command_buffer[256];
 
@@ -1697,7 +1820,7 @@ ui_event textui_get_command(int *count)
 
 	/* Get command */
 	while (1) {
-		/* Hack -- no flush needed */
+		/* No flush needed */
 		msg_flag = false;
 
 		/* Activate "command mode" */
@@ -1740,10 +1863,16 @@ ui_event textui_get_command(int *count)
 				}
 
 				case '^': {
-					char ch;
 					/* Allow "control chars" to be entered */
-					if (get_com("Control: ", &ch))
-						ke.key.code = KTRL(ch);
+					if (!get_com_ex("Control: ", &ke)
+							|| ke.type != EVT_KBRD) {
+						continue;
+					}
+					if (ENCODE_KTRL(ke.key.code)) {
+						ke.key.code = KTRL(ke.key.code);
+					} else {
+						ke.key.mods |= KC_MOD_CONTROL;
+					}
 					break;
 				}
 			}
@@ -1795,7 +1924,7 @@ bool key_confirm_command(unsigned char c)
 {
 	int i;
 
-	/* Hack -- Scan equipment */
+	/* Scan equipment */
 	for (i = 0; i < player->body.count; i++) {
 		char verify_inscrip[] = "^*";
 		unsigned n;
