@@ -13,8 +13,17 @@
  *  * Redistributions in binary form must reproduce the above copyright notice,
  *    this list of conditions and the following disclaimer in the documentation
  *    and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
+ * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
+ * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
+ * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
-#include "math.h"
 #include "angband.h"
 #include "cave.h"
 #include "cmds.h"
@@ -22,6 +31,7 @@
 #include "generate.h"
 #include "init.h"
 #include "mon-make.h"
+#include "mon-predicate.h"
 #include "monster.h"
 #include "obj-init.h"
 #include "obj-pile.h"
@@ -31,6 +41,7 @@
 #include "object.h"
 #include "ui-command.h"
 #include "wizard.h"
+#include <math.h>
 
 /**
  * The stats programs here will provide information on the dungeon, the monsters
@@ -427,7 +438,7 @@ static bool first_find(stat_first_find st)
 }
 
 /*
- * Add the number of drops for a specifci stat
+ * Add the number of drops for a specific stat
  */
 static void add_stats(stat_code st, bool vault, bool mon, int number)
 {
@@ -753,7 +764,7 @@ static void get_obj_data(const struct object *obj, int y, int x, bool mon,
 				add_stats(ST_RESTOREMANA_POTIONS, vault, mon, number);
 			} else if ((strstr(obj->kind->name, "Life")) ||
 					   (strstr(obj->kind->name, "*Healing*"))) {
-				add_stats(ST_ELVEN_RINGS, vault, mon, number);
+				add_stats(ST_BIGHEAL_POTIONS, vault, mon, number);
 			} else if (strstr(obj->kind->name, "Healing")) {
 				add_stats(ST_HEALING_POTIONS, vault, mon, number);
 			}
@@ -1092,8 +1103,7 @@ static void monster_death_stats(int m_idx)
 	assert(m_idx > 0);
 	mon = cave_monster(cave, m_idx);
 
-	/* Check if monster is UNIQUE */
-	uniq = rf_has(mon->race->flags,RF_UNIQUE);
+	uniq = monster_is_unique(mon);
 
 	/* Mimicked objects will have already been counted as floor objects */
 	mon->mimicked_obj = NULL;
@@ -1139,7 +1149,7 @@ static bool stats_monster(struct monster *mon, int i)
 	mon_total[lvl] += addval;
 
 	/* Increment unique count if appropriate */
-	if (rf_has(mon->race->flags, RF_UNIQUE)){
+	if (monster_is_unique(mon)) {
 
 		/* add to total */
 		uniq_total[lvl] += addval;
@@ -1157,8 +1167,7 @@ static bool stats_monster(struct monster *mon, int i)
 
 			mon_ood[lvl] += addval;
 
-			/* Is it a unique */
-			if (rf_has(mon->race->flags, RF_UNIQUE))
+			if (monster_is_unique(mon))
 				uniq_ood[lvl] += addval;
 	}
 
@@ -1168,8 +1177,7 @@ static bool stats_monster(struct monster *mon, int i)
 
 		mon_deadly[lvl] += addval;
 
-		/* Is it a unique? */
-		if (rf_has(mon->race->flags, RF_UNIQUE))
+		if (monster_is_unique(mon))
 			uniq_deadly[lvl] += addval;
 	}
 
@@ -1177,7 +1185,7 @@ static bool stats_monster(struct monster *mon, int i)
 	monster_death_stats(i);
 
 	/* remove the monster */
-	delete_monster_idx(i);
+	delete_monster_idx(cave, i);
 
 	/* success */
 	return true;
@@ -1185,7 +1193,7 @@ static bool stats_monster(struct monster *mon, int i)
 
 
 /**
- * Print heading infor for the file
+ * Print heading info for the file
  */
 static void print_heading(void)
 {
@@ -1229,6 +1237,12 @@ static void print_stats(int lvl)
 	file_putf(stats_log,"******** LEVEL %d , %d tries********* \n",lvl, tries);
 	file_putf(stats_log,"\n");
 
+	/* print gold info */
+	file_putf(stats_log," GOLD INFO \n");
+	file_putf(stats_log," Gold total: %f\n", gold_total[lvl]);
+	file_putf(stats_log," Gold monster: %f\n", gold_mon[lvl]);
+	file_putf(stats_log," Gold floor: %f\n", gold_floor[lvl]);
+
 	/* print monster heading */
 	file_putf(stats_log," MONSTER INFO \n");
 	file_putf(stats_log," Total monsters: %f OOD: %f Deadly: %f \n",
@@ -1268,28 +1282,21 @@ static void print_stats(int lvl)
  */
 static void mean_and_stdv(int array[TRIES_SIZE])
 {
-	int k, maxiter;
-	double tot = 0, mean, stdev, temp = 0;
+	int maxiter, iavg, ivar;
+	struct my_rational favg, fvar;
+	double avg, stdev;
 
 	/* Get the maximum iteration value */
 	maxiter = MIN(tries, TRIES_SIZE); 
 
-	/* Sum the array */
-	for (k = 0; k < maxiter; k++)
-		tot += array[k];
-
-	/* Compute the mean */
-	mean = tot / maxiter;
-
-	/* Sum up the squares */
-	for (k = 0; k < maxiter; k++) temp += (array[k] - mean) * (array[k] - mean);
-
-	/* Compute standard dev */
-	stdev = sqrt(temp / tries);
+	/* Get the statistics. */
+	iavg = mean(array, maxiter, &favg);
+	avg = (double)iavg + (double)favg.n / (double)favg.d;
+	ivar = variance(array, maxiter, false, false, &fvar);
+	stdev = sqrt((double)ivar + (double)fvar.n / (double)fvar.d);
 
 	/* Print to file */
-	file_putf(stats_log," mean: %f  std-dev: %f \n",mean,stdev);
-
+	file_putf(stats_log," mean: %f  std-dev: %f \n", avg, stdev);
 }
 
 /**
@@ -1366,7 +1373,7 @@ static void post_process_stats(void)
 	file_putf(stats_log,"80\t\t\t85\t\t\t90\t\t\t95\t\t\t100\n");
 	
 	for (i = 1; i < ST_FF_END; i++) {
-			file_putf(stats_log, stat_ff_message[i].name);
+			file_putf(stats_log, "%s", stat_ff_message[i].name);
 			prob_of_find(stat_all[stat_ff_message[i].st][0]);
 			mean_and_stdv(stat_ff_all[i]);
 	}
@@ -1455,7 +1462,7 @@ static void stats_collect_level(void)
  * This code will go through the artifact list and make each artifact
  * uncreated so that our sim player can find them again!
  */
-static void uncreate_artifacts(void)
+static void uncreate_all_artifacts(void)
 {
 	int i;
 
@@ -1518,7 +1525,7 @@ static void clearing_stats(void)
 	/* Do many iterations of the game */
 	for (iter = 0; iter < tries; iter++) {
 		/* Move all artifacts to uncreated */
-		uncreate_artifacts();
+		uncreate_all_artifacts();
 
 		/* Move all uniques to alive */
 		revive_uniques();
@@ -1736,10 +1743,10 @@ static void calc_cave_distances(int **cave_dist)
 					!square_isrubble(cave, loc(tx, ty))) continue;
 
 				/* Add the new location */
-				if (n_new == cap_new - 1) {
+				if (n_new == cap_new) {
 					cap_new *= 2;
 					ngrids = mem_realloc(ngrids,
-						cap_new * sizeof(ngrids));
+						cap_new * sizeof(*ngrids));
 				}
 				ngrids[n_new].y = ty;
 				ngrids[n_new].x = tx;
@@ -1770,47 +1777,169 @@ static void calc_cave_distances(int **cave_dist)
 /**
  * Generate several pits and collect statistics about the type of inhabitants.
  *
- * \param nsim Is the number of pits to generate.
+ * \param nsim Is the number of pits to generate at each depth.
  * \param pittype Must be 1 (pit), 2 (nest), or 3 (other).
- * \param depth Is the depth to use for the simulations.
+ * \param depth_min Is the minimum depth to use for the simulations.
+ * \param depth_max Is the maximum depth to use for the simulations.
+ *
+ * Does not account for failures that can occur in build_pit() or build_nest()
+ * when selecting the specific types of monsters to populate the room.
  */
-void pit_stats(int nsim, int pittype, int depth)
+void pit_stats(int nsim, int pittype, int depth_min, int depth_max)
 {
-	int *hist;
-	int j, p;
+	unsigned long *hist;
+	unsigned long *sum_hist;
+	const char *file_part;
+	char path[1024];
+	ang_file *pitfile;
+	int depth, p;
 
 	/* Initialize hist */
-	hist = mem_zalloc(z_info->pit_max * sizeof(*hist));
+	hist = mem_alloc(z_info->pit_max * sizeof(*hist));
+	if (depth_min < depth_max) {
+		sum_hist = mem_zalloc(z_info->pit_max * sizeof(*sum_hist));
+	} else {
+		sum_hist = NULL;
+	}
 
-	for (j = 0; j < nsim; j++) {
-		int i;
-		int pit_idx = 0;
-		int pit_dist = 999;
+	switch (pittype) {
+	case 1:
+		file_part = "pit_pit_stat.txt";
+		break;
 
-		for (i = 0; i < z_info->pit_max; i++) {
-			int offset, dist;
-			struct pit_profile *pit = &pit_info[i];
+	case 2:
+		file_part = "pit_nest_stat.txt";
+		break;
 
-			if (!pit->name || pit->room_type != pittype) continue;
+	case 3:
+		file_part = "pit_other_stat.txt";
+		break;
 
-			offset = Rand_normal(pit->ave, 10);
-			dist = ABS(offset - depth);
+	default:
+		file_part = NULL;
+	}
 
-			if (dist < pit_dist && one_in_(pit->rarity)) {
-				pit_idx = i;
-				pit_dist = dist;
+	if (file_part) {
+		path_build(path, sizeof(path), ANGBAND_DIR_USER, file_part);
+		pitfile = file_open(path, MODE_WRITE, FTYPE_TEXT);
+		if (pitfile) {
+			(void)file_putf(pitfile,
+				"%d simulations at each depth\n", nsim);
+			(void)file_putf(pitfile, "depth");
+			for (p = 0; p < z_info->pit_max; ++p) {
+				const struct pit_profile *pit = &pit_info[p];
+
+				if (pit->name) {
+					/*
+					 * Replace spaces or tabs in the name
+					 * with underscores so columns in the
+					 * output can be extracted by splitting
+					 * on whitespace.
+					 */
+					char *name_copy =
+						string_make(pit->name);
+					char *c = name_copy;
+
+					while (*c) {
+						if (*c == ' ' || *c == '\t') {
+							*c = '_';
+						}
+						++c;
+					}
+
+					(void)file_putf(pitfile,
+						"\t%s", name_copy);
+					string_free(name_copy);
+				}
 			}
+			(void)file_putf(pitfile, "\n");
+		}
+	} else {
+		pitfile = NULL;
+	}
+
+	for (depth = depth_min; depth <= depth_max; ++depth) {
+		int j;
+
+		for (p = 0; p < z_info->pit_max; ++p) {
+			hist[p] = 0;
 		}
 
-		hist[pit_idx]++;
+		for (j = 0; j < nsim; j++) {
+			int i;
+			int pit_idx = 0;
+			int pit_dist = 999;
+
+			for (i = 0; i < z_info->pit_max; i++) {
+				int offset, dist;
+				const struct pit_profile *pit = &pit_info[i];
+
+				if (!pit->name || pit->room_type != pittype) {
+					continue;
+				}
+
+				offset = Rand_normal(pit->ave, 10);
+				dist = ABS(offset - depth);
+
+				if (dist < pit_dist && one_in_(pit->rarity)) {
+					pit_idx = i;
+					pit_dist = dist;
+				}
+			}
+
+			hist[pit_idx]++;
+		}
+
+		if (pitfile) {
+			(void)file_putf(pitfile, "%d", depth);
+			for (p = 0; p < z_info->pit_max; ++p) {
+				const struct pit_profile *pit = &pit_info[p];
+
+				if (pit->name) {
+					(void)file_putf(pitfile, "\t%lu",
+						hist[p]);
+				}
+			}
+			(void)file_putf(pitfile, "\n");
+		}
+
+		if (sum_hist) {
+			for (p = 0; p < z_info->pit_max; ++p) {
+				sum_hist[p] += hist[p];
+			}
+		}
 	}
 
-	for (p = 0; p < z_info->pit_max; p++) {
-		struct pit_profile *pit = &pit_info[p];
-		if (pit->name)
-			msg("Type: %s, Number: %d.", pit->name, hist[p]);
+	for (p = 0; p < z_info->pit_max; ++p) {
+		const struct pit_profile *pit = &pit_info[p];
+
+		if (pit->name) {
+			msg("Type %s: %lu.", pit->name, (sum_hist) ?
+				sum_hist[p] : hist[p]);
+		}
 	}
 
+	if (pitfile) {
+		if (sum_hist) {
+			(void)file_putf(pitfile, "\nsum");
+			for (p = 0; p < z_info->pit_max; ++p) {
+				const struct pit_profile *pit = &pit_info[p];
+
+				if (pit->name) {
+					(void)file_putf(pitfile, "\t%lu",
+						sum_hist[p]);
+				}
+			}
+			(void)file_putf(pitfile, "\n");
+		}
+		if (file_close(pitfile)) {
+			msg("Results are also recorded in %s.", file_part);
+		}
+	}
+
+	if (sum_hist) {
+		mem_free(sum_hist);
+	}
 	mem_free(hist);
 
 	return;
@@ -1993,7 +2122,7 @@ static double stddev_d_sum_sum2(struct d_sum_sum2 s, int count)
 
 struct tunnel_aggregate {
 	/*
-	 * Hold the sums for for the normalized number of steps, number of
+	 * Hold the sums for the normalized number of steps, number of
 	 * wall piercings, normalized number of grids excavated, normalized
 	 * starting distance, and normalized final distance.  The first
 	 * includes all tunnels, the second only those that had early
@@ -2487,7 +2616,7 @@ static void dump_generation_stats(ang_file *fo, const struct cgen_stats *gs)
 	int i;
 
 	file_put(fo, "Number of Successful Levels::\n");
-	file_putf(fo, "%d\n\n", (unsigned long) gs->nsuccess);
+	file_putf(fo, "%d\n\n", gs->nsuccess);
 
 	file_put(fo, "Level Builder Success Count, Probability, and Failure Rate Per Successful Level::\n");
 	for (i = 0; i < z_info->profile_max; ++i) {
@@ -2678,7 +2807,7 @@ static void dump_generation_stats(ang_file *fo, const struct cgen_stats *gs)
 
 		file_putf(fo, "\"%s\" Tunneling Total Number, Success Rate, Early Termination Rate::\n", name);
 		file_putf(fo, "%lu\t%.6f\t%.6f\t%.6f\t%.6f\n\n",
-			gs->ta[i].cv_all.count,
+			(unsigned long) gs->ta[i].cv_all.count,
 			gs->ta[i].success_frac.sum /
 			gs->level_counts[0][i], stddev_d_sum_sum2(
 			gs->ta[i].success_frac, gs->level_counts[0][i]),
@@ -2961,7 +3090,7 @@ void disconnect_stats(int nsim, bool stop_on_disconnect)
 {
 }
 
-void pit_stats(int nsim, int pittype, int depth)
+void pit_stats(int nsim, int pittype, int depth_min, int depth_max)
 {
 }
 #endif /* USE_STATS */
@@ -3077,6 +3206,12 @@ static bool is_easily_traversed(struct chunk *c, struct loc grid)
 		square_isrubble(c, grid);
 }
 
+static bool is_floor_trap(struct chunk *c, struct loc grid)
+{
+	/* Using square_istrap by itself will include locked doors. */
+	return square_istrap(c, grid) && !square_iscloseddoor(c, grid);
+}
+
 static bool is_impassable_rubble(struct chunk *c, struct loc grid)
 {
 	return !square_ispassable(c, grid) && square_isrubble(c, grid);
@@ -3119,7 +3254,7 @@ void stat_grid_counter_simple(struct chunk *c, struct grid_counts counts[3])
 		{ square_isfloor, 0, 0, 0 },
 		{ square_isupstairs, 0, 0, 0 },
 		{ square_isdownstairs, 0, 0, 0 },
-		{ square_istrap, 0, 0, 0 },
+		{ is_floor_trap, 0, 0, 0 },
 		{ square_isfiery, 0, 0, 0 },
 		{ is_impassable_rubble, 0, 0, 0 },
 		{ is_passable_rubble, 0, 0, 0 },
